/*========== Includes ==========*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <dirent.h>
#include <math.h>
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
#include <windows.h>
#undef PlaySound
#endif

#include <SDL.h>
#include <SDL_audio.h>
#include <SDL_thread.h>
#include <SDL_ttf.h>
#include <SDL_image.h>
/*========== Includes ==========*/

/*========== Defines ==========*/
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
#define SLASH "\\"
#define HERE ""
#define BATCH_FILE "playtest.bat"
#else
#define SLASH "/"
#define HERE "./"
#define BATCH_FILE "playtest.sh"
#endif

#define PNG_VARIOUS "png" SLASH "various" SLASH
#define PNG_BUTTONS "png" SLASH "buttons" SLASH
#define PNG_GAMEPAD "png" SLASH "gamepad" SLASH
#define PNG_LIVING "png" SLASH "living" SLASH
#define PNG_SLIVING "png" SLASH "sliving" SLASH
#define PNG_EXTRAS "png" SLASH "extras" SLASH
#define PNG_ROOMS "png" SLASH "rooms" SLASH
#define PNG_L0003 "png" SLASH "l0003" SLASH

#define EXIT_NORMAL 0
#define EXIT_ERROR 1
#define EDITOR_NAME "TODO"
#define EDITOR_VERSION "TODO"
#define COPYRIGHT "TODO"
#define DISKA "fdi" SLASH "DISK_A.FDI"
#define MAX_FILENAME 100
#define MAX_DATA 720
#define ENTRIES 256
#define MAX_FILEDATA 120000 /*** CHTAB6.DAT is 119324. ***/
#define MAX_OPTION 100
#define MAX_LEVEL 15
#define ROOMS 24
#define TILES 30
#define EVENTS 256
#define BLOCKS 128
#define WINDOW_WIDTH 690
#define WINDOW_HEIGHT 467
#define DIR_FILES "files"
#define MAX_IMG 200
#define MAX_CON 30
#define NUM_SOUNDS 20 /*** Sounds that may play at the same time. ***/
#define BAR_FULL 443
#define MAX_TEXT 100
#define REFRESH 25 /*** That is 40 frames per second, 1000/25. ***/
#define ADJ_BASE_X 415
#define ADJ_BASE_Y 65

#ifndef O_BINARY
#define O_BINARY 0
#endif
/*========== Defines ==========*/

int iFdA;
int iDebug;
int iNoAudio;
int iScale;
int iFullscreen;
int iStartLevel;
int iNoController;
int iBrokenRoomLinks;
int iDone[24 + 2];
int iHeader;
int iNoAnim;
int iCurRoom;
SDL_Window *window;
unsigned int iWindowID;
SDL_Renderer *ascreen;
unsigned int iActiveWindowID;
SDL_Cursor *curArrow;
SDL_Cursor *curWait;
SDL_Cursor *curHand;
int iHourglassFrame;
int iSandFrame;
int iPreLoaded;
int iNrToPreLoad;
int iCurrentBarHeight;
TTF_Font *font11;
TTF_Font *font15;
TTF_Font *font20;
int iStatusBarFrame;
int iScreen;
int iXPos, iYPos;
int iChanged;
int iCurLevel;
int iCurGuard;
int iDownAt;
int iSelected;
int iExtras;
int iLastBlock;
int iLastModif;
int iOnTile;
int iCloseOn;
int iChangeEvent;
int iGuardSkill;
int iModified;
int iPlaytest;
int iMovingRoom;
int iMovingNewBusy;
int iChangingBrokenRoom;
int iChangingBrokenSide;
int arRoomMoving[ROOMS + 1 + 2][ROOMS + 2];
int iMovingOldX;
int iMovingOldY;
int iMovingNewX;
int iMovingNewY;
int iCustomBlock;
int iCustomHover, iCustomHoverOld;

char sFileNames[ENTRIES + 2][MAX_FILENAME + 2];
int iFirstSectors[ENTRIES + 2];

unsigned char sFileCom[MAX_FILEDATA + 2];
unsigned char sFileDecom[MAX_FILEDATA + 2];
int iSizeCom;
int iSizeDecom;
int iPosIn;

/*** level ***/
int sBytesA[2048 + 2];
int sBytesB[256 + 2];
int sBytesC[256 + 2];
int arBlockFlag[BLOCKS + 2];
int arBlockObject[BLOCKS + 2];
int arBlock[ROOMS + 2][TILES + 2];
int arModif[ROOMS + 2][TILES + 2];
int arEventRoom[EVENTS + 2];
int arEventTile[EVENTS + 2];
int arEventNext[EVENTS + 2];
int arRoomLinks[ROOMS + 2][4 + 2];
int sBytesD[1 + 2];
int sBytesE[24 + 2];
int sBytesF[24 + 2];
int sBytesG[15 + 2];
int iKidRoom, iKidPos, iKidDir;
int sBytesH[4 + 2];
int arGuardLocations[ROOMS + 2];
int arGuardDirections[ROOMS + 2];
int arGuardUnused1[ROOMS + 2];
int arGuardUnused2[ROOMS + 2];
int arGuardSkills[ROOMS + 2];
int arGuardUnused3[ROOMS + 2];
int arGuardUnused4[ROOMS + 2];
int sBytesI[17 + 2];

int arRoomLinksBroken[ROOMS + 2][4 + 2];
int iMinX, iMaxX, iMinY, iMaxY, iStartRoomsX, iStartRoomsY;

unsigned int gamespeed;
Uint32 looptime;

static const int broken_room_x = 431;
static const int broken_room_y = 81;
static const int broken_side_x[5] = { 0, 416, 446, 431, 431 };
static const int broken_side_y[5] = { 0, 81, 81, 66, 96 };

/*** controller ***/
int iController;
SDL_GameController *controller;
char sControllerName[MAX_CON + 2];
SDL_Joystick *joystick;
SDL_Haptic *haptic;

/*** for text ***/
SDL_Surface *message;
SDL_Texture *messaget;
SDL_Rect offset;
SDL_Color color_bl = {0x00, 0x00, 0x00, 255};
SDL_Color color_wh = {0xff, 0xff, 0xff, 255};

/* All files in disk A.
 * 0 = no compression / 1 = compressed
 */
static const char *files[][2] = {
	{"", ""},
	{"MUSIC.SYS", "0"},
	{"MAIN.PRG", "0"},
	{"SUBPROG.PRG", "0"},
	{"BJLOGO.PRG", "0"},
	{"PALET.DAT", "0"},
	{"FONT.DAT", "0"},
	{"FIRE.DAT", "0"},
	{"MUSIC1.MUS", "0"},
	{"MUSIC2.MUS", "0"},
	{"MUSIC3.MUS", "0"},
	{"MUSIC4.MUS", "0"},
	{"FIGHT.MUS", "0"},
	{"FIGHTF.MUS", "0"},
	{"FIGHTS.MUS", "0"},
	{"FIGHTK.MUS", "0"},
	{"VICT1.MUS", "0"},
	{"APPEAR.MUS", "0"},
	{"APPEAR2.MUS", "0"},
	{"PICKUP.MUS", "0"},
	{"POWUP.MUS", "0"},
	{"KUSURI.MUS", "0"},
	{"CLEAR.MUS", "0"},
	{"EXIT.MUS", "0"},
	{"DEAD.MUS", "0"},
	{"DEAD2.MUS", "0"},
	{"OPEN.MUS", "0"},
	{"ROOM1.MUS", "0"},
	{"ROOM2.MUS", "0"},
	{"END1.MUS", "0"},
	{"END2.MUS", "0"},
	{"END3.MUS", "0"},
	{"TIMEOUT.MUS", "0"},
	{"MENU.MUS", "0"},
	{"SOUND.EFC", "0"},
	{"BEEP.EFC", "0"},
	{"JAFFER1.DAT", "0"},
	{"JAFFER2.DAT", "0"},
	{"JAFFER3.DAT", "0"},
	{"OPEN1.DAT", "0"},
	{"OPEN2.DAT", "0"},
	{"OPEN3.DAT", "0"},
	{"OPEN11.DAT", "0"},
	{"OPEN12.DAT", "0"},
	{"OPEN32.DAT", "0"},
	{"OPEN33.DAT", "0"},
	{"JAF_S1.DAT", "0"},
	{"JAF_S2.DAT", "0"},
	{"JAF_E1.DAT", "0"},
	{"JAF_E2.DAT", "0"},
	{"JAF_E3.DAT", "0"},
	{"JAF_E4.DAT", "0"},
	{"TITLE.DAT", "0"},
	{"TITLE2.DAT", "0"},
	{"TITLE3.DAT", "0"},
	{"TITLE4.DAT", "0"},
	{"PROOM.DAT", "0"},
	{"VCLIP.DAT", "0"},
	{"MOYOU.DAT", "0"},
	{"TRAP.DAT", "1"},
	{"PLATE1.DAT", "1"},
	{"PLATE4.DAT", "1"},
	{"PLATE7.DAT", "1"},
	{"PLATEA.DAT", "1"},
	{"PLATEC.DAT", "1"},
	{"PLATEE.DAT", "1"},
	{"CHTAB1.DAT", "1"},
	{"CHTAB2.DAT", "1"},
	{"CHTAB3.DAT", "1"},
	{"CHTAB4F.DAT", "1"},
	{"CHTAB4G.DAT", "1"},
	{"CHTAB4K.DAT", "1"},
	{"CHTAB4S.DAT", "1"},
	{"CHTAB4V.DAT", "1"},
	{"CHTAB4X.DAT", "1"},
	{"CHTAB5.DAT", "1"},
	{"CHTAB6.DAT", "1"},
	{"CHTAB7.DAT", "1"},
	{"LEV01.CHR", "1"},
	{"LEV04.CHR", "1"},
	{"LEV07.CHR", "1"},
	{"LEV10.CHR", "1"},
	{"LEV12.CHR", "1"},
	{"LEV14.CHR", "1"},
	{"LEV15.CHR", "1"},
	{"LEV00.MAP", "1"},
	{"LEV01.MAP", "1"},
	{"LEV02.MAP", "1"},
	{"LEV03.MAP", "1"},
	{"LEV04.MAP", "1"},
	{"LEV05.MAP", "1"},
	{"LEV06.MAP", "1"},
	{"LEV07.MAP", "1"},
	{"LEV08.MAP", "1"},
	{"LEV09.MAP", "1"},
	{"LEV10.MAP", "1"},
	{"LEV11.MAP", "1"},
	{"LEV12.MAP", "1"},
	{"LEV13.MAP", "1"},
	{"LEV14.MAP", "1"},
	{"LEV15.MAP", "1"},
	{"DEMOPLAY.KEY", "0"},
	{"EFONT.DAT", "0"},
	{"ENDING1.DAT", "0"},
	{"ENDING2.DAT", "0"},
	{"ENDING3.DAT", "0"},
	{"ENDING4.DAT", "0"},
	{"ENDING5.DAT", "0"},
	{"ENDING6.DAT", "0"}
};

SDL_Texture *imgloading;
SDL_Texture *imghourglasssprite;
SDL_Texture *imgsandsprite;
SDL_Texture *imgblack;
SDL_Texture *imgleft_0, *imgleft_1;
SDL_Texture *imgright_0, *imgright_1;
SDL_Texture *imgup_0, *imgup_1;
SDL_Texture *imgdown_0, *imgdown_1;
SDL_Texture *imglrno, *imgudno, *imgudnonfo;
SDL_Texture *imgroomsoff, *imgroomson_0, *imgroomson_1;
SDL_Texture *imgbroomsoff, *imgbroomson_0, *imgbroomson_1;
SDL_Texture *imgeventsoff, *imgeventson_0, *imgeventson_1;
SDL_Texture *imgsaveoff, *imgsaveon_0, *imgsaveon_1;
SDL_Texture *imgquit_0, *imgquit_1;
SDL_Texture *imgprevoff, *imgprevon_0, *imgprevon_1;
SDL_Texture *imgnextoff, *imgnexton_0, *imgnexton_1;
SDL_Texture *imgkidr[2 + 2], *imgkidl[2 + 2];
SDL_Texture *imgbar;
SDL_Texture *extras[10 + 2];
SDL_Texture *imgl0003t, *imgl0406t, *imgl0709t, *imgl1011t;
SDL_Texture *imgl1213t, *imgl14t, *imgl15t;
SDL_Texture *imgclosebig_0, *imgclosebig_1;
SDL_Texture *imgbordersl;
SDL_Texture *imgrl, *imgbrl;
SDL_Texture *imgroom[ROOMS + 2];
SDL_Texture *imgsrc, *imgsrs, *imgsrm, *imgsrp, *imgsrb;
SDL_Texture *imgevents, *imgsele, *imgeventu;
SDL_Texture *imgunk;
SDL_Texture *imgplaytest;
SDL_Texture *imgl0003[255 + 2];
SDL_Texture *imgselblock;
SDL_Texture *imgchover;

/*** guards ***/
SDL_Texture *imgguardbg[2 + 2];
SDL_Texture *imgguardpb[2 + 2];
SDL_Texture *imgguardunk[2 + 2];
SDL_Texture *imgskel[2 + 2];
SDL_Texture *imgfat[2 + 2];
SDL_Texture *imgjaffar[2 + 2];
SDL_Texture *imgshadow[2 + 2];
/***/
SDL_Texture *imgguardsel[2 + 2];
SDL_Texture *imgskelsel[2 + 2];
SDL_Texture *imgfatsel[2 + 2];
SDL_Texture *imgjaffarsel[2 + 2];
SDL_Texture *imgshadowsel[2 + 2];

struct sample {
	Uint8 *data;
	Uint32 dpos;
	Uint32 dlen;
} sounds[NUM_SOUNDS];

void ShowUsage (void);
void GetOptionValue (char *sArgv, char *sValue);
void LoadFileCom (int iFd, char *sFileName);
void LSeek (int iFd, int iOffset);
int ReadFromFile (int iFd, char *sWhat, int iSize, unsigned char *sRetString);
void WriteCharByChar (int iFd, unsigned char *sString, int iLength);
void PrIfDe (char *sString);
void LoadFileNamesAndFirstSectors (int iFd);
void DecompressFile (char *sFileName);
void DecompFun (int iType, int cPos1, int cPos2, int cPos3, int cPos4);
void CompressFile (char *sFileName);
int CheckHeader (int iFd, int iOffset);
unsigned char NibbleLower (unsigned char cChar);
unsigned char NibbleUpper (unsigned char cChar);
unsigned char NibbleMerge (unsigned char cUpper, unsigned char cLower);
void SaveFileCom (char *sFileName);
void SaveFileDecom (char *sFileName);
void LoadLevel (int iLevel);
void SaveLevel (int iLevel);
void InitScreenAction (char *sAction);
void InitScreen (void);
void Quit (void);
void GetBlockName (int iBlock, int iModif, char *sName);
void GetFlagName (int iFlag, char *sName);
void GetObjectName (int iObject, char *sName);
char cShowDirection (int iDirection);
void LoadFonts (void);
void MixAudio (void *unused, Uint8 *stream, int iLen);
void PreLoad (char *sPath, char *sPNG, SDL_Texture **imgImage);
void ShowImage (SDL_Texture *img, int iX, int iY, char *sImageInfo,
	SDL_Renderer *screen, float fMultiply, int iXYScale);
void LoadingBar (int iBarHeight);
void ShowScreen (int iScreenS);
void InitPopUp (void);
void InitPopUpSave (void);
void ShowTile (int iBlock, int iLocation);
void LocationToXY (int iLocation, int iXY[4]);
int InArea (int iUpperLeftX, int iUpperLeftY,
	int iLowerRightX, int iLowerRightY);
void PlaySound (char *sFile);
int BrokenRoomLinks (int iPrint);
void CheckSides (int iRoom, int iX, int iY);
void Prev (void);
void Next (void);
void DisplayText (int iStartX, int iStartY, int iFontSize,
	char arText[9 + 2][MAX_TEXT + 2], int iLines, TTF_Font *font,
	SDL_Color back, SDL_Color fore);
void CustomRenderCopy (SDL_Texture* src, char *sImageInfo, SDL_Rect* srcrect,
	SDL_Renderer* dst, SDL_Rect *dstrect);
void ChangeEvent (int iAmount, int iChangePos);
int OnLevelBar (void);
void InitRooms (void);
void WhereToStart (void);
void ShowRooms (int iRoom, int iX, int iY, int iNext);
void CenterNumber (SDL_Renderer *screen, int iNumber, int iX, int iY,
	SDL_Color fore, int iHex);
int EventInfo (int iNr, int iType);
int BitFromByte (int iByte, int iGetBitNr);
int BitsToInt (char *sString);
void EventRoom (int iRoom);
void EventDoor (int iTile);
void EventNext (int iNoNext);
void ClearRoom (void);
void UseTile (int iTile, int iLocation, int iRoom);
int MouseSelectAdj (void);
void ChangePos (int iLocation);
void ShowChange (int iLocation);
void RemoveOldRoom (void);
void AddNewRoom (int iX, int iY, int iRoom);
void LinkPlus (void);
void LinkMinus (void);
void CopyPaste (int iRoom, int iAction);
void FlipRoom (int iRoom, int iAxis);
void RunLevel (int iLevel);
int StartGame (void *unused);
void ModifyForPlaytest (int iLevel);
void ModifyBack (void);
void Zoom (int iToggleFull);
void Help (void);
void Sprinkle (void);
void EXE (void);
void SetLocation (int iRoom, int iTile, int iBlock, int iModif);
void CreateDir (char *sDir);
void ExtractAllFiles (void);
void LoadFileToDecom (char *sFileName);
void ApplySkillIfNecessary (int iLoc);
void ChangeCustom (int iAmount);
int OnTile (void);
int IsDisabled (int iTile);

/*****************************************************************************/
int main (int argc, char *argv[])
/*****************************************************************************/
{
	char sStartLevel[MAX_OPTION + 2];
	time_t tm;
	SDL_version verc, verl;

	/*** Used for looping. ***/
	int iLoopArg;

	iDebug = 0;
	iNoAudio = 0;
	iScale = 1;
	iFullscreen = 0;
	iStartLevel = 1;
	iNoController = 0;
	iExtras = 0;
	iLastBlock = 0;
	iLastModif = 0;
	iNoAnim = 0;
	iModified = 0;

	if (argc > 1)
	{
		for (iLoopArg = 1; iLoopArg <= argc - 1; iLoopArg++)
		{
			if ((strcmp (argv[iLoopArg], "-h") == 0) ||
				(strcmp (argv[iLoopArg], "-?") == 0) ||
				(strcmp (argv[iLoopArg], "--help") == 0))
			{
				ShowUsage();
			}
			else if ((strcmp (argv[iLoopArg], "-v") == 0) ||
				(strcmp (argv[iLoopArg], "--version") == 0))
			{
				printf ("%s %s\n", EDITOR_NAME, EDITOR_VERSION);
				exit (EXIT_NORMAL);
			}
			else if ((strcmp (argv[iLoopArg], "-d") == 0) ||
				(strcmp (argv[iLoopArg], "--debug") == 0))
			{
				iDebug = 1;
			}
			else if ((strcmp (argv[iLoopArg], "-n") == 0) ||
				(strcmp (argv[iLoopArg], "--noaudio") == 0))
			{
				iNoAudio = 1;
			}
			else if ((strcmp (argv[iLoopArg], "-z") == 0) ||
				(strcmp (argv[iLoopArg], "--zoom") == 0))
			{
				iScale = 2;
			}
			else if ((strcmp (argv[iLoopArg], "-f") == 0) ||
				(strcmp (argv[iLoopArg], "--fullscreen") == 0))
			{
				iFullscreen = SDL_WINDOW_FULLSCREEN_DESKTOP;
			}
			else if ((strncmp (argv[iLoopArg], "-l=", 3) == 0) ||
				(strncmp (argv[iLoopArg], "--level=", 8) == 0))
			{
				GetOptionValue (argv[iLoopArg], sStartLevel);
				iStartLevel = atoi (sStartLevel);
				if ((iStartLevel < 0) || (iStartLevel > MAX_LEVEL))
				{
					iStartLevel = 1;
				}
			}
			else if ((strcmp (argv[iLoopArg], "-s") == 0) ||
				(strcmp (argv[iLoopArg], "--static") == 0))
			{
				iNoAnim = 1;
			}
			else if ((strcmp (argv[iLoopArg], "-k") == 0) ||
				(strcmp (argv[iLoopArg], "--keyboard") == 0))
			{
				iNoController = 1;
			}
			else
			{
				ShowUsage();
			}
		}
	}

	iFdA = open (DISKA, O_RDONLY|O_BINARY);
	if (iFdA == -1)
	{
		printf ("[FAILED] Could not open \"%s\": %s!\n",
			DISKA, strerror (errno));
		exit (EXIT_ERROR);
	}

	/*** iHeader ***/
	if (CheckHeader (iFdA, 0x00) == 1)
	{
		iHeader = 0;
	} else if (CheckHeader (iFdA, 0x1000) == 1) {
		iHeader = 1;
	} else {
		printf ("[FAILED] Cannot find header bytes!\n");
		exit (EXIT_ERROR);
	}

	ExtractAllFiles();

	srand ((unsigned)time(&tm));

	LoadLevel (iStartLevel);

	/*** Show the SDL version used for compiling and linking. ***/
	if (iDebug == 1)
	{
		SDL_VERSION (&verc);
		SDL_GetVersion (&verl);
		printf ("[ INFO ] Compiled with SDL %u.%u.%u, linked with SDL %u.%u.%u.\n",
			verc.major, verc.minor, verc.patch, verl.major, verl.minor, verl.patch);
	}

	InitScreen();

	Quit();

	return 0;
}
/*****************************************************************************/
void ShowUsage (void)
/*****************************************************************************/
{
	printf ("%s %s\n%s\n\n", EDITOR_NAME, EDITOR_VERSION, COPYRIGHT);
	printf ("Usage:\n");
	printf ("  %s [OPTIONS]\n\nOptions:\n", EDITOR_NAME);
	printf ("  -h, -?,    --help           display this help and exit\n");
	printf ("  -v,        --version        output version information and"
		" exit\n");
	printf ("  -d,        --debug          also show levels on the console\n");
	printf ("  -n,        --noaudio        do not play sound effects\n");
	printf ("  -z,        --zoom           double the interface size\n");
	printf ("  -f,        --fullscreen     start in fullscreen mode\n");
	printf ("  -l=NR,     --level=NR       start in level NR\n");
	printf ("  -s,        --static         do not display animations\n");
	printf ("  -k,        --keyboard       do not use a game controller\n");
	printf ("\n");
	exit (EXIT_NORMAL);
}
/*****************************************************************************/
void GetOptionValue (char *sArgv, char *sValue)
/*****************************************************************************/
{
	int iTemp;
	char sTemp[MAX_OPTION + 2];

	iTemp = strlen (sArgv) - 1;
	snprintf (sValue, MAX_OPTION, "%s", "");
	while (sArgv[iTemp] != '=')
	{
		snprintf (sTemp, MAX_OPTION, "%c%s", sArgv[iTemp], sValue);
		snprintf (sValue, MAX_OPTION, "%s", sTemp);
		iTemp--;
	}
}
/*****************************************************************************/
void LoadFileCom (int iFd, char *sFileName)
/*****************************************************************************/
{
	int iCurrentSector, iNextSector;
	int iExit;
	int iOffset;
	unsigned char sNextSector[2 + 2];
	int iSize;
	unsigned char sSectorData[0x400 + 2];

	/*** Used for looping. ***/
	int iLoopEntry;
	int iLoopByte;

	sFileCom[0] = '\0';
	iSizeCom = 0;

	/*** iCurrentSector ***/
	for (iLoopEntry = 1; iLoopEntry <= ENTRIES; iLoopEntry++)
	{
		if (strcmp (sFileNames[iLoopEntry], sFileName) == 0)
			{ iCurrentSector = iFirstSectors[iLoopEntry]; }
	}
	if (iDebug == 1)
		{ printf ("[ INFO ] Current sector: %i\n", iCurrentSector); }

	iExit = 0;
	do {
		if ((iCurrentSector < 0x0008) || (iCurrentSector > 0x04CF))
		{
			printf ("[ WARN ] Invalid sector: %i!\n", iCurrentSector);
		}

		/*** iNextSector ***/
		iOffset = 0x400 + (iCurrentSector * 2);
		LSeek (iFd, iOffset);
		ReadFromFile (iFd, "Next Sector", 2, sNextSector);
		iNextSector = sNextSector[0] + (sNextSector[1] << 8);
		if (iDebug == 1)
			{ printf ("[ INFO ] Next sector: %i\n", iNextSector); }

		iOffset = iCurrentSector * 0x400;
		LSeek (iFd, iOffset);
		ReadFromFile (iFd, "Sector Data", 0x400, sSectorData);
		if (iNextSector < 0xFC00)
		{
			iSize = 0x400; /*** Everything. ***/
			for (iLoopByte = 0; iLoopByte < iSize; iLoopByte++)
			{
				sFileCom[iSizeCom] = sSectorData[iLoopByte];
				iSizeCom++;
			}
			iCurrentSector = iNextSector;
		} else {
			iSize = iNextSector - 0xFC00;
			if (iSize == 0) { iSize = 0x400; }
			for (iLoopByte = 0; iLoopByte < iSize; iLoopByte++)
			{
				sFileCom[iSizeCom] = sSectorData[iLoopByte];
				iSizeCom++;
			}
			iExit = 1;
		}
	} while (iExit == 0);
	sFileCom[iSizeCom] = '\0';
}
/*****************************************************************************/
void LSeek (int iFd, int iOffset)
/*****************************************************************************/
{
	lseek (iFd, iOffset + (iHeader * 0x1000), SEEK_SET);
}
/*****************************************************************************/
int ReadFromFile (int iFd, char *sWhat, int iSize, unsigned char *sRetString)
/*****************************************************************************/
{
	int iLength;
	int iRead;
	char sRead[1 + 2];
	int iEOF;

	if ((iDebug == 1) && (strcmp (sWhat, "") != 0))
	{
		printf ("[  OK  ] Loading: %s\n", sWhat);
	}
	iLength = 0;
	iEOF = 0;
	do {
		iRead = read (iFd, sRead, 1);
		switch (iRead)
		{
			case -1:
				printf ("[FAILED] Could not read: %s!\n", strerror (errno));
				exit (EXIT_ERROR);
				break;
			case 0: PrIfDe ("[ INFO ] End of file\n"); iEOF = 1; break;
			default:
				sRetString[iLength] = sRead[0];
				iLength++;
				break;
		}
	} while ((iLength < iSize) && (iEOF == 0));
	sRetString[iLength] = '\0';

	return (iLength);
}
/*****************************************************************************/
void WriteCharByChar (int iFd, unsigned char *sString, int iLength)
/*****************************************************************************/
{
	/*** Used for looping. ***/
	int iPos;

	for (iPos = 0; iPos < iLength; iPos++)
	{
		write (iFd, (char *)&sString[iPos], 1);
	}
}
/*****************************************************************************/
void PrIfDe (char *sString)
/*****************************************************************************/
{
	if (iDebug == 1) { printf ("%s", sString); }
}
/*****************************************************************************/
void LoadFileNamesAndFirstSectors (int iFd)
/*****************************************************************************/
{
	unsigned char sBaseName[8 + 2];
	unsigned char sExtension[3 + 2];
	unsigned char sUnknown[3 + 2];
	unsigned char sFirstSector[2 + 2];

	/*** Used for looping. ***/
	int iLoopEntry;
	int iLoopChar;

	LSeek (iFd, 0x1000);
	for (iLoopEntry = 1; iLoopEntry <= ENTRIES; iLoopEntry++)
	{
		ReadFromFile (iFd, "Base Name", 8, sBaseName);
		for (iLoopChar = 7; iLoopChar > 0; iLoopChar--)
			{ if (sBaseName[iLoopChar] == ' ') { sBaseName[iLoopChar] = '\0'; } }
		ReadFromFile (iFd, "Extension", 3, sExtension);
		snprintf (sFileNames[iLoopEntry], MAX_FILENAME,
			"%s.%s", sBaseName, sExtension);
		ReadFromFile (iFd, "", 3, sUnknown);
		ReadFromFile (iFd, "First Sector", 2, sFirstSector);
		iFirstSectors[iLoopEntry] = sFirstSector[0] + (sFirstSector[1] << 8);

		if (iDebug == 1)
		{
			if (iFirstSectors[iLoopEntry] != 65535)
			{
				printf ("[ INFO ] First sector %s: %i\n",
					sFileNames[iLoopEntry], iFirstSectors[iLoopEntry]);
			} else {
				printf ("[ INFO ] (Unused entry.)\n");
			}
		}
	}
}
/*****************************************************************************/
void DecompressFile (char *sFileName)
/*****************************************************************************/
{
	/*** This function decompresses sFileCom into sFileDecom. ***/

	unsigned char sArg[4 + 2];
	unsigned char cN, cA, cB, cC, cD;
	unsigned char cLast1, cLast2, cLast3, cLast4;
	unsigned char cPos1, cPos2, cPos3, cPos4;
	int iTimes;

	/*** Used for looping. ***/
	int iLoopRepeat;

	sFileDecom[0] = '\0';
	iSizeDecom = 0;

	iPosIn = 0;
	do {
		if (iDebug == 1)
		{
			printf ("[ INFO ] (%s) Compression at %i: 0x%02X\n",
				sFileName, iPosIn, sFileCom[iPosIn]);
		}
		switch (sFileCom[iPosIn])
		{
			case 0x00:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				sArg[4] = sFileCom[iPosIn + 4];
				iPosIn+=5;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[3];
				sFileDecom[iSizeDecom + 3] = sArg[4];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[3], sArg[4]); }
				iSizeDecom+=4;
				break;
			case 0x01:
				/*** No arguments. ***/
				iPosIn+=1;
				/***/
				sFileDecom[iSizeDecom + 0] = sFileDecom[iSizeDecom - 4];
				sFileDecom[iSizeDecom + 1] = sFileDecom[iSizeDecom - 3];
				sFileDecom[iSizeDecom + 2] = sFileDecom[iSizeDecom - 2];
				sFileDecom[iSizeDecom + 3] = sFileDecom[iSizeDecom - 1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sFileDecom[iSizeDecom - 4], sFileDecom[iSizeDecom - 3],
					sFileDecom[iSizeDecom - 2], sFileDecom[iSizeDecom - 1]); }
				iSizeDecom+=4;
				break;
			case 0x11:
				sArg[1] = sFileCom[iPosIn + 1];
				iPosIn+=2;
				/***/
				cLast1 = sFileDecom[iSizeDecom - 4]; // TODO: maybe move into for()
				cLast2 = sFileDecom[iSizeDecom - 3];
				cLast3 = sFileDecom[iSizeDecom - 2];
				cLast4 = sFileDecom[iSizeDecom - 1];
				iTimes = sArg[1] + 1;
				for (iLoopRepeat = 1; iLoopRepeat <= iTimes; iLoopRepeat++)
				{
					sFileDecom[iSizeDecom + 0] = cLast1;
					sFileDecom[iSizeDecom + 1] = cLast2;
					sFileDecom[iSizeDecom + 2] = cLast3;
					sFileDecom[iSizeDecom + 3] = cLast4;
					if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
						cLast1, cLast2, cLast3, cLast4); }
					iSizeDecom+=4;
				}
				break;
			case 0x21:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				cLast1 = sFileDecom[iSizeDecom - 4]; // TODO: maybe move into for()
				cLast2 = sFileDecom[iSizeDecom - 3];
				cLast3 = sFileDecom[iSizeDecom - 2];
				cLast4 = sFileDecom[iSizeDecom - 1];
				iTimes = sArg[1] + (sArg[2] << 8) + 1;
				for (iLoopRepeat = 1; iLoopRepeat <= iTimes; iLoopRepeat++)
				{
					sFileDecom[iSizeDecom + 0] = cLast1;
					sFileDecom[iSizeDecom + 1] = cLast2;
					sFileDecom[iSizeDecom + 2] = cLast3;
					sFileDecom[iSizeDecom + 3] = cLast4;
					if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
						cLast1, cLast2, cLast3, cLast4); }
					iSizeDecom+=4;
				}
				break;
			case 0x81:
				/*** No arguments. ***/
				iPosIn+=1;
				/***/
				sFileDecom[iSizeDecom + 0] = sFileDecom[iSizeDecom - 8];
				sFileDecom[iSizeDecom + 1] = sFileDecom[iSizeDecom - 7];
				sFileDecom[iSizeDecom + 2] = sFileDecom[iSizeDecom - 6];
				sFileDecom[iSizeDecom + 3] = sFileDecom[iSizeDecom - 5];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sFileDecom[iSizeDecom - 8], sFileDecom[iSizeDecom - 7],
					sFileDecom[iSizeDecom - 6], sFileDecom[iSizeDecom - 5]); }
				iSizeDecom+=4;
				break;
			case 0x91:
				sArg[1] = sFileCom[iPosIn + 1];
				iPosIn+=2;
				/***/
				cLast1 = sFileDecom[iSizeDecom - 8]; // TODO: maybe move into for()
				cLast2 = sFileDecom[iSizeDecom - 7];
				cLast3 = sFileDecom[iSizeDecom - 6];
				cLast4 = sFileDecom[iSizeDecom - 5];
				iTimes = sArg[1] + 1;
				for (iLoopRepeat = 1; iLoopRepeat <= iTimes; iLoopRepeat++)
				{
					sFileDecom[iSizeDecom + 0] = cLast1;
					sFileDecom[iSizeDecom + 1] = cLast2;
					sFileDecom[iSizeDecom + 2] = cLast3;
					sFileDecom[iSizeDecom + 3] = cLast4;
					if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
						cLast1, cLast2, cLast3, cLast4); }
					iSizeDecom+=4;
				}
				break;
			case 0x02:
				sArg[1] = sFileCom[iPosIn + 1];
				iPosIn+=2;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[1];
				sFileDecom[iSizeDecom + 2] = sArg[1];
				sFileDecom[iSizeDecom + 3] = sArg[1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[1], sArg[1], sArg[1]); }
				iSizeDecom+=4;
				break;
			case 0x03:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[2];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[2], sArg[2]); }
				iSizeDecom+=4;
				break;
			case 0x13:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[1];
				sFileDecom[iSizeDecom + 3] = sArg[1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[1], sArg[1]); }
				iSizeDecom+=4;
				break;
			case 0x23:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[1];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[1], sArg[2], sArg[1]); }
				iSizeDecom+=4;
				break;
			case 0x33:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[1];
				sFileDecom[iSizeDecom + 2] = sArg[1];
				sFileDecom[iSizeDecom + 3] = sArg[2];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[1], sArg[1], sArg[2]); }
				iSizeDecom+=4;
				break;
			case 0x04:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[1];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[2];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[1], sArg[2], sArg[2]); }
				iSizeDecom+=4;
				break;
			case 0x14:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[1];
				sFileDecom[iSizeDecom + 3] = sArg[2];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[1], sArg[2]); }
				iSizeDecom+=4;
				break;
			case 0x24:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[2], sArg[1]); }
				iSizeDecom+=4;
				break;
			case 0x44:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[1];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[3];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[1], sArg[2], sArg[3]); }
				iSizeDecom+=4;
				break;
			case 0x54:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[1];
				sFileDecom[iSizeDecom + 3] = sArg[3];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[1], sArg[3]); }
				iSizeDecom+=4;
				break;
			case 0x64:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[3];
				sFileDecom[iSizeDecom + 3] = sArg[1];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[3], sArg[1]); }
				iSizeDecom+=4;
				break;
			case 0x74:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[2];
				sFileDecom[iSizeDecom + 3] = sArg[3];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[2], sArg[3]); }
				iSizeDecom+=4;
				break;
			case 0x84:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[3];
				sFileDecom[iSizeDecom + 3] = sArg[2];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[3], sArg[2]); }
				iSizeDecom+=4;
				break;
			case 0x94:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				sArg[3] = sFileCom[iPosIn + 3];
				iPosIn+=4;
				/***/
				sFileDecom[iSizeDecom + 0] = sArg[1];
				sFileDecom[iSizeDecom + 1] = sArg[2];
				sFileDecom[iSizeDecom + 2] = sArg[3];
				sFileDecom[iSizeDecom + 3] = sArg[3];
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					sArg[1], sArg[2], sArg[3], sArg[3]); }
				iSizeDecom+=4;
				break;
			case 0x05: case 0x15: case 0x25: case 0x35: case 0x45: case 0x55:
			case 0x65: case 0x75: case 0x85: case 0x95: case 0xA5: case 0xB5:
			case 0xC5: case 0xD5: case 0xE5: case 0xF5:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				cN = NibbleUpper (sFileCom[iPosIn - 3]);
				cA = NibbleLower (sArg[1]);
				cB = NibbleUpper (sArg[1]);
				cC = NibbleLower (sArg[2]);
				cD = NibbleUpper (sArg[2]);
				sFileDecom[iSizeDecom + 0] = NibbleMerge (cN, cA);
				sFileDecom[iSizeDecom + 1] = NibbleMerge (cN, cB);
				sFileDecom[iSizeDecom + 2] = NibbleMerge (cN, cC);
				sFileDecom[iSizeDecom + 3] = NibbleMerge (cN, cD);
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					NibbleMerge (cN, cA), NibbleMerge (cN, cB),
					NibbleMerge (cN, cC), NibbleMerge (cN, cD)); }
				iSizeDecom+=4;
				break;
			case 0x06: case 0x16: case 0x26: case 0x36: case 0x46: case 0x56:
			case 0x66: case 0x76: case 0x86: case 0x96: case 0xA6: case 0xB6:
			case 0xC6: case 0xD6: case 0xE6: case 0xF6:
				sArg[1] = sFileCom[iPosIn + 1];
				sArg[2] = sFileCom[iPosIn + 2];
				iPosIn+=3;
				/***/
				cN = NibbleUpper (sFileCom[iPosIn - 3]);
				cA = NibbleLower (sArg[1]);
				cB = NibbleUpper (sArg[1]);
				cC = NibbleLower (sArg[2]);
				cD = NibbleUpper (sArg[2]);
				sFileDecom[iSizeDecom + 0] = NibbleMerge (cA, cN);
				sFileDecom[iSizeDecom + 1] = NibbleMerge (cB, cN);
				sFileDecom[iSizeDecom + 2] = NibbleMerge (cC, cN);
				sFileDecom[iSizeDecom + 3] = NibbleMerge (cD, cN);
				if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
					NibbleMerge (cA, cN), NibbleMerge (cB, cN),
					NibbleMerge (cC, cN), NibbleMerge (cD, cN)); }
				iSizeDecom+=4;
				break;
			case 0x07: case 0x17: case 0x27: case 0x37: case 0x47: case 0x57:
			case 0x67: case 0x77: case 0x87: case 0x97: case 0xA7: case 0xB7:
			case 0xC7: case 0xD7: /*** No 0xE7. ***/ case 0xF7:
				DecompFun (sFileCom[iPosIn] - 7, 0x00, 0x00, 0x00, 0x00);
				break;
			case 0x08: case 0x18: case 0x28: case 0x38: case 0x48: case 0x58:
			case 0x68: case 0x78: case 0x88: case 0x98: case 0xA8: case 0xB8:
			case 0xC8: case 0xD8: /*** No 0xE8. ***/ case 0xF8:
				DecompFun (sFileCom[iPosIn] - 8, 0xFF, 0xFF, 0xFF, 0xFF);
				break;
			case 0x09: case 0x19: case 0x29: case 0x39: case 0x49: case 0x59:
			case 0x69: case 0x79: case 0x89: case 0x99: case 0xA9: case 0xB9:
			case 0xC9: case 0xD9: /*** No 0xE9 or 0xF9. ***/
				cPos1 = sFileDecom[iSizeDecom - 4];
				cPos2 = sFileDecom[iSizeDecom - 3];
				cPos3 = sFileDecom[iSizeDecom - 2];
				cPos4 = sFileDecom[iSizeDecom - 1];
				DecompFun (sFileCom[iPosIn] - 9, cPos1, cPos2,
					cPos3, cPos4);
				break;
			case 0x0A: case 0x1A: case 0x2A: case 0x3A: case 0x4A: case 0x5A:
			case 0x6A: case 0x7A: case 0x8A: case 0x9A: case 0xAA: case 0xBA:
			case 0xCA: case 0xDA: /*** No 0xEA or 0xFA. ***/
				cPos1 = sFileDecom[iSizeDecom - 8];
				cPos2 = sFileDecom[iSizeDecom - 7];
				cPos3 = sFileDecom[iSizeDecom - 6];
				cPos4 = sFileDecom[iSizeDecom - 5];
				DecompFun (sFileCom[iPosIn] - 10, cPos1, cPos2,
					cPos3, cPos4);
				break;
			default:
				printf ("[FAILED] Invalid compression. Head byte at pos. %i/%i:"
					" 0x%02X!\n", iPosIn, iSizeCom, sFileCom[iPosIn]);
				exit (EXIT_ERROR);
				break;
		}
	} while (iPosIn < iSizeCom);
	sFileDecom[iSizeDecom] = '\0';
}
/*****************************************************************************/
void DecompFun (int iType, int cPos1, int cPos2, int cPos3, int cPos4)
/*****************************************************************************/
{
	unsigned char sArg[4 + 2];

	switch (iType)
	{
		case 0x00:
			sArg[1] = sFileCom[iPosIn + 1];
			iPosIn+=2;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = sArg[1];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, cPos2, cPos3, sArg[1]); }
			iSizeDecom+=4;
			break;
		case 0x10:
			sArg[1] = sFileCom[iPosIn + 1];
			iPosIn+=2;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = sArg[1];
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, cPos2, sArg[1], cPos4); }
			iSizeDecom+=4;
			break;
		case 0x20:
			sArg[1] = sFileCom[iPosIn + 1];
			iPosIn+=2;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = sArg[1];
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, sArg[1], cPos3, cPos4); }
			iSizeDecom+=4;
			break;
		case 0x30:
			sArg[1] = sFileCom[iPosIn + 1];
			iPosIn+=2;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], cPos2, cPos3, cPos4); }
			iSizeDecom+=4;
			break;
		case 0x40:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = sArg[1];
			sFileDecom[iSizeDecom + 3] = sArg[2];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, cPos2, sArg[1], sArg[2]); }
			iSizeDecom+=4;
			break;
		case 0x50:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = sArg[1];
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = sArg[2];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, sArg[1], cPos3, sArg[2]); }
			iSizeDecom+=4;
			break;
		case 0x60:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = sArg[1];
			sFileDecom[iSizeDecom + 2] = sArg[2];
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, sArg[1], sArg[2], cPos4); }
			iSizeDecom+=4;
			break;
		case 0x70:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = sArg[2];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], cPos2, cPos3, sArg[2]); }
			iSizeDecom+=4;
			break;
		case 0x80:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = sArg[2];
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], cPos2, sArg[2], cPos4); }
			iSizeDecom+=4;
			break;
		case 0x90:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			iPosIn+=3;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = sArg[2];
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], sArg[2], cPos3, cPos4); }
			iSizeDecom+=4;
			break;
		case 0xA0:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			sArg[3] = sFileCom[iPosIn + 3];
			iPosIn+=4;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = sArg[1];
			sFileDecom[iSizeDecom + 2] = sArg[2];
			sFileDecom[iSizeDecom + 3] = sArg[3];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, sArg[1], sArg[2], sArg[3]); }
			iSizeDecom+=4;
			break;
		case 0xB0:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			sArg[3] = sFileCom[iPosIn + 3];
			iPosIn+=4;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = sArg[2];
			sFileDecom[iSizeDecom + 3] = sArg[3];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], cPos2, sArg[2], sArg[3]); }
			iSizeDecom+=4;
			break;
		case 0xC0:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			sArg[3] = sFileCom[iPosIn + 3];
			iPosIn+=4;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = sArg[2];
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = sArg[3];
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], sArg[2], cPos3, sArg[3]); }
			iSizeDecom+=4;
			break;
		case 0xD0:
			sArg[1] = sFileCom[iPosIn + 1];
			sArg[2] = sFileCom[iPosIn + 2];
			sArg[3] = sFileCom[iPosIn + 3];
			iPosIn+=4;
			/***/
			sFileDecom[iSizeDecom + 0] = sArg[1];
			sFileDecom[iSizeDecom + 1] = sArg[2];
			sFileDecom[iSizeDecom + 2] = sArg[3];
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				sArg[1], sArg[2], sArg[3], cPos4); }
			iSizeDecom+=4;
			break;
		case 0xF0:
			/*** No arguments. ***/
			iPosIn+=1;
			/***/
			sFileDecom[iSizeDecom + 0] = cPos1;
			sFileDecom[iSizeDecom + 1] = cPos2;
			sFileDecom[iSizeDecom + 2] = cPos3;
			sFileDecom[iSizeDecom + 3] = cPos4;
			if (iDebug == 1) { printf ("[ INFO ] %02X, %02X, %02X, %02X\n",
				cPos1, cPos2, cPos3, cPos4); }
			iSizeDecom+=4;
			break;
	}
}
/*****************************************************************************/
void CompressFile (char *sFileName)
/*****************************************************************************/
{
	// TODO Not yet in use, neither tested. Will be used in SaveLevel().

	/*** This function compresses sFileDecom into sFileCom. ***/

	/*** Used for looping. ***/
	int iLoopByte;

	if (iDebug == 1)
	{
		printf ("[ INFO ] (%s) Compressing.\n", sFileName);
	}

	// TODO improve compression
	iSizeCom = 0;
	for (iLoopByte = 0; iLoopByte <= iSizeDecom; iLoopByte+=4)
	{
		sFileCom[iSizeCom] = 0x00;
		sFileCom[iSizeCom + 1] = sFileDecom[iLoopByte];
		sFileCom[iSizeCom + 2] = sFileDecom[iLoopByte + 1];
		sFileCom[iSizeCom + 3] = sFileDecom[iLoopByte + 2];
		sFileCom[iSizeCom + 4] = sFileDecom[iLoopByte + 3];
		iSizeCom+=5;
	}
}
/*****************************************************************************/
int CheckHeader (int iFd, int iOffset)
/*****************************************************************************/
{
	unsigned char sCheckHeader[2 + 2];

	LSeek (iFd, iOffset);
	ReadFromFile (iFd, "Check Header", 2, sCheckHeader);
	if ((sCheckHeader[0] == 0xEB) && (sCheckHeader[1] == 0x0A))
		{ return (1); } else { return (0); }
}
/*****************************************************************************/
unsigned char NibbleLower (unsigned char cChar)
/*****************************************************************************/
{
	return (cChar & 0x0f); /*** 0f = 00001111 as mask ***/
}
/*****************************************************************************/
unsigned char NibbleUpper (unsigned char cChar)
/*****************************************************************************/
{
	return ((cChar & 0xf0) >> 4); /*** f0 = 11110000 as mask ***/
}
/*****************************************************************************/
unsigned char NibbleMerge (unsigned char cUpper, unsigned char cLower)
/*****************************************************************************/
{
	return ((cUpper << 4) | cLower);
}
/*****************************************************************************/
void SaveFileCom (char *sFileName)
/*****************************************************************************/
{
	/* Merely saves sFileCom to sFileName.
	 * Does NOT compress.
	 */

	int iFdD;

	iFdD = open (sFileName, O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, 0600);

	WriteCharByChar (iFdD, sFileCom, iSizeCom);

	close (iFdD);
}
/*****************************************************************************/
void SaveFileDecom (char *sFileName)
/*****************************************************************************/
{
	/* Merely saves sFileDecom to sFileName.
	 * Does NOT decompress.
	 */

	int iFdD;

	iFdD = open (sFileName, O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, 0600);

	WriteCharByChar (iFdD, sFileDecom, iSizeDecom);

	close (iFdD);
}
/*****************************************************************************/
void LoadLevel (int iLevel)
/*****************************************************************************/
{
	char sFileName[MAX_FILENAME + 2];
	int iPos;
	int iBlock, iModif;
	int iLeft, iRight, iUp, iDown;
	char sName[7 + 2];
	int iFlagObject, iFlag, iObject;
	int iEventByte1, iEventByte2;
	char sBinary[50 + 2];

	/*** Used for looping. ***/
	int iLoopByte;
	int iLoopBlock;
	int iLoopRoom;
	int iLoopTile;
	int iLoopMinus;
	int iLoopEvent;

	iChanged = 0; /*** TODO correct location? ***/
	snprintf (sFileName, MAX_FILENAME, "%s%sLEV%02i.MAP",
		DIR_FILES, SLASH, iLevel);
	LoadFileToDecom (sFileName);

	iPos = 0;
	for (iLoopByte = 1; iLoopByte <= 2048; iLoopByte++)
		{ sBytesA[iLoopByte] = sFileDecom[iPos + (iLoopByte - 1)]; }
	iPos+=2048;
	for (iLoopByte = 1; iLoopByte <= 256; iLoopByte++)
		{ sBytesB[iLoopByte] = sFileDecom[iPos + (iLoopByte - 1)]; }
	iPos+=256;
	for (iLoopByte = 1; iLoopByte <= 256; iLoopByte++)
		{ sBytesC[iLoopByte] = sFileDecom[iPos + (iLoopByte - 1)]; }
	iPos+=256;
	PrIfDe ("[  OK  ] Loading: Flags and Objects\n");
	for (iLoopBlock = 0; iLoopBlock < BLOCKS; iLoopBlock++)
	{
		iFlagObject = sFileDecom[iPos + (iLoopBlock * 2)];
		iFlag = iFlagObject & 0x07;
		iObject = iFlagObject & 0xf8;
		arBlockFlag[iLoopBlock] = iFlag;
		arBlockObject[iLoopBlock] = iObject;
		if (iDebug == 1)
		{
			printf ("[ INFO ] Block %i: flag %i, object %i\n",
				iLoopBlock, iFlag, iObject);
		}
	}
	iPos+=256;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		if (iDebug == 1) { printf ("\nRoom: %i\n\n", iLoopRoom); }
		for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
		{
			iBlock = sFileDecom[iPos + ((iLoopRoom - 1) * 30) + (iLoopTile - 1)];
			iModif = sFileDecom[iPos + 720 + ((iLoopRoom - 1) * 30) +
				(iLoopTile - 1)];
			arBlock[iLoopRoom][iLoopTile] = iBlock;
			arModif[iLoopRoom][iLoopTile] = iModif;
			if (iDebug == 1)
			{
				GetBlockName (iBlock, iModif, sName);
				printf ("%s", sName);
				if ((iLoopTile != 10) && (iLoopTile != 20) && (iLoopTile != 30))
					{ printf ("|"); }
				if ((iLoopTile == 10) || (iLoopTile == 20))
				{
					printf ("\n");
					for (iLoopMinus = 1; iLoopMinus <= 79; iLoopMinus++)
						{ printf ("-"); }
					printf ("\n");
				}
			}
		}
		if (iDebug == 1) { printf ("\n\n"); }
	}
	iPos+=(2 * ROOMS * TILES);
	for (iLoopEvent = 1; iLoopEvent <= EVENTS; iLoopEvent++)
	{
		iEventByte1 = sFileDecom[iPos + (iLoopEvent - 1)];
		iEventByte2 = sFileDecom[iPos + EVENTS + (iLoopEvent - 1)];
		snprintf (sBinary, 50, "%i%i%i%i%i",
			BitFromByte (iEventByte2, 7),
			BitFromByte (iEventByte2, 6),
			BitFromByte (iEventByte2, 5),
			BitFromByte (iEventByte1, 6),
			BitFromByte (iEventByte1, 5));
		arEventRoom[iLoopEvent] = BitsToInt (sBinary);
		snprintf (sBinary, 50, "%i%i%i%i%i",
			BitFromByte (iEventByte1, 4),
			BitFromByte (iEventByte1, 3),
			BitFromByte (iEventByte1, 2),
			BitFromByte (iEventByte1, 1),
			BitFromByte (iEventByte1, 0));
		arEventTile[iLoopEvent] = BitsToInt (sBinary) + 1;
		snprintf (sBinary, 50, "%i",
			BitFromByte (iEventByte1, 7));
		arEventNext[iLoopEvent] = BitsToInt (sBinary) ^ 1;
		if (iDebug == 1)
		{
			printf ("[ INFO ] Event %i changes the door in room: %i, location:"
				" %i.",
				iLoopEvent,
				EventInfo (iLoopEvent, 1),
				EventInfo (iLoopEvent, 2));
			switch (EventInfo (iLoopEvent, 3))
			{
				case 1: printf (" (next: yes)\n"); break;
				case 0: printf (" (next: no)\n"); break;
			}
		}
	}
	iPos+=(2 * EVENTS);
	PrIfDe ("[  OK  ] Loading: Room Links\n");
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		iLeft = sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 0];
		iRight = sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 1];
		iUp = sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 2];
		iDown = sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 3];
		arRoomLinks[iLoopRoom][1] = iLeft;
		arRoomLinks[iLoopRoom][2] = iRight;
		arRoomLinks[iLoopRoom][3] = iUp;
		arRoomLinks[iLoopRoom][4] = iDown;
		if (iDebug == 1)
		{
			printf ("[ INFO ] Room %i is connected to room (0 = none):"
				" l%i r%i u%i d%i\n", iLoopRoom, iLeft, iRight, iUp, iDown);
		}
	}
	iPos+=96;
	sBytesD[1] = sFileDecom[iPos];
	iPos+=1;
	for (iLoopByte = 1; iLoopByte <= 24; iLoopByte++)
		{ sBytesE[iLoopByte] = sFileDecom[iPos + (iLoopByte - 1)]; }
	iPos+=24;
	for (iLoopByte = 1; iLoopByte <= 24; iLoopByte++)
		{ sBytesF[iLoopByte] = sFileDecom[iPos + (iLoopByte - 1)]; }
	iPos+=24;
	for (iLoopByte = 1; iLoopByte <= 15; iLoopByte++)
		{ sBytesG[iLoopByte] = sFileDecom[iPos + (iLoopByte - 1)]; }
	iPos+=15;
	iKidRoom = sFileDecom[iPos];
	iKidPos = sFileDecom[iPos + 1] + 1;
	iKidDir = sFileDecom[iPos + 2];
	if ((iLevel == 1) || (iLevel == 13))
	{
		if (iKidDir == 0x00) { iKidDir = 0xFF; } else { iKidDir = 0x00; }
	}
	if (iDebug == 1)
	{
		printf ("[ INFO ] The kid starts in room: %i, position: %i, turned: "
			"%c\n", iKidRoom, iKidPos, cShowDirection (iKidDir));
	}
	iPos+=3;
	for (iLoopByte = 1; iLoopByte <= 4; iLoopByte++)
		{ sBytesH[iLoopByte] = sFileDecom[iPos + (iLoopByte - 1)]; }
	iPos+=4;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		arGuardLocations[iLoopRoom] = sFileDecom[iPos + (iLoopRoom - 1)] + 1;
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		arGuardDirections[iLoopRoom] = sFileDecom[iPos + (iLoopRoom - 1)];
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		arGuardUnused1[iLoopRoom] = sFileDecom[iPos + (iLoopRoom - 1)];
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		arGuardUnused2[iLoopRoom] = sFileDecom[iPos + (iLoopRoom - 1)];
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		arGuardSkills[iLoopRoom] = sFileDecom[iPos + (iLoopRoom - 1)];
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		arGuardUnused3[iLoopRoom] = sFileDecom[iPos + (iLoopRoom - 1)];
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		arGuardUnused4[iLoopRoom] = sFileDecom[iPos + (iLoopRoom - 1)];
	}
	iPos+=24;
	if (iDebug == 1)
	{
		for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
		{
			if (arGuardLocations[iLoopRoom] != 31)
			{
				printf ("[ INFO ] A guard in room: %i, location: %i, turned: %c, "
					"skill: %i\n", iLoopRoom, arGuardLocations[iLoopRoom],
					cShowDirection (arGuardDirections[iLoopRoom]),
					arGuardSkills[iLoopRoom]);
				printf ("[ INFO ] Extra data around the above guard: %i|%i|%i|%i\n",
					arGuardUnused1[iLoopRoom], arGuardUnused2[iLoopRoom],
					arGuardUnused3[iLoopRoom], arGuardUnused4[iLoopRoom]);
			}
		}
	}
	for (iLoopByte = 1; iLoopByte <= 17; iLoopByte++)
		{ sBytesI[iLoopByte] = sFileDecom[iPos + (iLoopByte - 1)]; }

	PrIfDe ("[  OK  ] Checking for broken room links.\n");
	iBrokenRoomLinks = BrokenRoomLinks (1);

	switch (iLevel)
	{
		case 0: iCurGuard = 0; break;
		case 1: iCurGuard = 0; break;
		case 2: iCurGuard = 0; break;
		case 3: iCurGuard = 2; break;
		case 4: iCurGuard = 0; break;
		case 5: iCurGuard = 0; break;
		case 6: iCurGuard = 1; break;
		case 7: iCurGuard = 0; break;
		case 8: iCurGuard = 0; break;
		case 9: iCurGuard = 0; break;
		case 10: iCurGuard = 0; break;
		case 11: iCurGuard = 0; break;
		case 12: iCurGuard = 4; break; /*** TODO shadow ***/
		case 13: iCurGuard = 3; break; /*** TODO jaffar ***/
		case 14: iCurGuard = 65535; break;
		case 15: iCurGuard = 65535; break;
	}

	iCurLevel = iLevel;
	iCurRoom = iKidRoom;
	iOnTile = 1;
	iPlaytest = 0;
}
/*****************************************************************************/
void SaveLevel (int iLevel)
/*****************************************************************************/
{
	char sFileName[MAX_FILENAME + 2];
	int iPos;
	int iBlock, iModif;
	int iLeft, iRight, iUp, iDown;
	int iFlagObject, iFlag, iObject;
	int iEventByte1, iEventByte2;
	char sBinary[50 + 2];

	/*** Used for looping. ***/
	int iLoopByte;
	int iLoopBlock;
	int iLoopRoom;
	int iLoopTile;
	int iLoopEvent;

	if (iChanged == 0) { return; }

	iPos = 0;
	for (iLoopByte = 1; iLoopByte <= 2048; iLoopByte++)
		{ sFileDecom[iPos + (iLoopByte - 1)] = sBytesA[iLoopByte]; }
	iPos+=2048;
	for (iLoopByte = 1; iLoopByte <= 256; iLoopByte++)
		{ sFileDecom[iPos + (iLoopByte - 1)] = sBytesB[iLoopByte]; }
	iPos+=256;
	for (iLoopByte = 1; iLoopByte <= 256; iLoopByte++)
		{ sFileDecom[iPos + (iLoopByte - 1)] = sBytesC[iLoopByte]; }
	iPos+=256;
	for (iLoopBlock = 0; iLoopBlock < BLOCKS; iLoopBlock++)
	{
		iFlag = arBlockFlag[iLoopBlock];
		iObject = arBlockObject[iLoopBlock];
		iFlagObject = iObject + iFlag;
		sFileDecom[iPos + (iLoopBlock * 2)] = iFlagObject;
	}
	iPos+=256;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
		{
			iBlock = arBlock[iLoopRoom][iLoopTile];
			iModif = arModif[iLoopRoom][iLoopTile];
			sFileDecom[iPos + ((iLoopRoom - 1) * 30) + (iLoopTile - 1)]
				= iBlock;
			sFileDecom[iPos + 720 + ((iLoopRoom - 1) * 30) + (iLoopTile - 1)]
				= iModif;
		}
	}
	iPos+=(2 * ROOMS * TILES);
	for (iLoopEvent = 1; iLoopEvent <= EVENTS; iLoopEvent++)
	{
		snprintf (sBinary, 50, "%i%i%i%i%i%i%i%i",
			arEventNext[iLoopEvent] ^ 1,
			BitFromByte (arEventRoom[iLoopEvent], 1),
			BitFromByte (arEventRoom[iLoopEvent], 0),
			BitFromByte (arEventTile[iLoopEvent] - 1, 4),
			BitFromByte (arEventTile[iLoopEvent] - 1, 3),
			BitFromByte (arEventTile[iLoopEvent] - 1, 2),
			BitFromByte (arEventTile[iLoopEvent] - 1, 1),
			BitFromByte (arEventTile[iLoopEvent] - 1, 0));
		iEventByte1 = BitsToInt (sBinary);
		snprintf (sBinary, 50, "%i%i%i%i%i%i%i%i",
			BitFromByte (arEventRoom[iLoopEvent], 4),
			BitFromByte (arEventRoom[iLoopEvent], 3),
			BitFromByte (arEventRoom[iLoopEvent], 2),
			0,
			0,
			0,
			0,
			0);
		iEventByte2 = BitsToInt (sBinary);
		sFileDecom[iPos + (iLoopEvent - 1)] = iEventByte1;
		sFileDecom[iPos + EVENTS + (iLoopEvent - 1)] = iEventByte2;
	}
	iPos+=(2 * EVENTS);
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		iLeft = arRoomLinks[iLoopRoom][1];
		iRight = arRoomLinks[iLoopRoom][2];
		iUp = arRoomLinks[iLoopRoom][3];
		iDown = arRoomLinks[iLoopRoom][4];
		sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 0] = iLeft;
		sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 1] = iRight;
		sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 2] = iUp;
		sFileDecom[iPos + ((iLoopRoom - 1) * 4) + 3] = iDown;
	}
	iPos+=96;
	sFileDecom[iPos] = sBytesD[1];
	iPos+=1;
	for (iLoopByte = 1; iLoopByte <= 24; iLoopByte++)
		{ sFileDecom[iPos + (iLoopByte - 1)] = sBytesE[iLoopByte]; }
	iPos+=24;
	for (iLoopByte = 1; iLoopByte <= 24; iLoopByte++)
		{ sFileDecom[iPos + (iLoopByte - 1)] = sBytesF[iLoopByte]; }
	iPos+=24;
	for (iLoopByte = 1; iLoopByte <= 15; iLoopByte++)
		{ sFileDecom[iPos + (iLoopByte - 1)] = sBytesG[iLoopByte]; }
	iPos+=15;
	sFileDecom[iPos] = iKidRoom;
	sFileDecom[iPos + 1] = iKidPos - 1;
	sFileDecom[iPos + 2] = iKidDir;
	if ((iLevel == 1) || (iLevel == 13))
	{
		if (sFileDecom[iPos + 2] == 0x00)
			{ sFileDecom[iPos + 2] = 0xFF; }
				else { sFileDecom[iPos + 2] = 0x00; }
	}
	iPos+=3;
	for (iLoopByte = 1; iLoopByte <= 4; iLoopByte++)
		{ sFileDecom[iPos + (iLoopByte - 1)] = sBytesH[iLoopByte]; }
	iPos+=4;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		sFileDecom[iPos + (iLoopRoom - 1)] = arGuardLocations[iLoopRoom] - 1;
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		sFileDecom[iPos + (iLoopRoom - 1)] = arGuardDirections[iLoopRoom];
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		sFileDecom[iPos + (iLoopRoom - 1)] = arGuardUnused1[iLoopRoom];
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		sFileDecom[iPos + (iLoopRoom - 1)] = arGuardUnused2[iLoopRoom];
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		sFileDecom[iPos + (iLoopRoom - 1)] = arGuardSkills[iLoopRoom];
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		sFileDecom[iPos + (iLoopRoom - 1)] = arGuardUnused3[iLoopRoom];
	}
	iPos+=24;
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		sFileDecom[iPos + (iLoopRoom - 1)] = arGuardUnused4[iLoopRoom];
	}
	iPos+=24;
	for (iLoopByte = 1; iLoopByte <= 17; iLoopByte++)
		{ sFileDecom[iPos + (iLoopByte - 1)] = sBytesI[iLoopByte]; }

	snprintf (sFileName, MAX_FILENAME, "%s%sLEV%02i.MAP",
		DIR_FILES, SLASH, iLevel);
	SaveFileDecom (sFileName);

	// TODO
	// Use files[] and CompressFile() to put together DISKA.
	// This includes a new directory with updated first sectors, and
	// the file allocation table.

	PlaySound ("wav/save.wav");

	iChanged = 0;
}
/*****************************************************************************/
void InitScreenAction (char *sAction)
/*****************************************************************************/
{
	int iEventDoor;
	int iEventRoom;

	if (strcmp (sAction, "left") == 0)
	{
		switch (iScreen)
		{
			case 1:
				iSelected--;
				switch (iSelected)
				{
					case 0: iSelected = 10; break;
					case 10: iSelected = 20; break;
					case 20: iSelected = 30; break;
				}
				break;
			case 2:
				if (iBrokenRoomLinks == 0)
				{
					if (iMovingRoom != 0)
					{
						if (iMovingNewX != 1) { iMovingNewX--; }
							else { iMovingNewX = 25; }
						PlaySound ("wav/cross.wav");
					}
				} else {
					if (iChangingBrokenSide != 1)
					{
						iChangingBrokenSide = 1;
					} else {
						switch (iChangingBrokenRoom)
						{
							case 1: iChangingBrokenRoom = 4; break;
							case 5: iChangingBrokenRoom = 8; break;
							case 9: iChangingBrokenRoom = 12; break;
							case 13: iChangingBrokenRoom = 16; break;
							case 17: iChangingBrokenRoom = 20; break;
							case 21: iChangingBrokenRoom = 24; break;
							default: iChangingBrokenRoom--; break;
						}
					}
				}
				break;
			case 3:
				iEventDoor = EventInfo (iChangeEvent, 2);
				switch (iEventDoor)
				{
					case 1: iEventDoor = 10; break;
					case 11: iEventDoor = 20; break;
					case 21: iEventDoor = 30; break;
					default: iEventDoor--; break;
				}
				EventDoor (iEventDoor);
				break;
		}
	}

	if (strcmp (sAction, "right") == 0)
	{
		switch (iScreen)
		{
			case 1:
				iSelected++;
				switch (iSelected)
				{
					case 11: iSelected = 1; break;
					case 21: iSelected = 11; break;
					case 31: iSelected = 21; break;
				}
				break;
			case 2:
				if (iBrokenRoomLinks == 0)
				{
					if (iMovingRoom != 0)
					{
						if (iMovingNewX != 25) { iMovingNewX++; }
							else { iMovingNewX = 1; }
						PlaySound ("wav/cross.wav");
					}
				} else {
					if (iChangingBrokenSide != 2)
					{
						iChangingBrokenSide = 2;
					} else {
						switch (iChangingBrokenRoom)
						{
							case 4: iChangingBrokenRoom = 1; break;
							case 8: iChangingBrokenRoom = 5; break;
							case 12: iChangingBrokenRoom = 9; break;
							case 16: iChangingBrokenRoom = 13; break;
							case 20: iChangingBrokenRoom = 17; break;
							case 24: iChangingBrokenRoom = 21; break;
							default: iChangingBrokenRoom++; break;
						}
					}
				}
				break;
			case 3:
				iEventDoor = EventInfo (iChangeEvent, 2);
				switch (iEventDoor)
				{
					case 10: iEventDoor = 1; break;
					case 20: iEventDoor = 11; break;
					case 30: iEventDoor = 21; break;
					default: iEventDoor++; break;
				}
				EventDoor (iEventDoor);
				break;
		}
	}

	if (strcmp (sAction, "up") == 0)
	{
		switch (iScreen)
		{
			case 1:
				if (iSelected > 10) { iSelected-=10; }
					else { iSelected+=20; }
				break;
			case 2:
				if (iBrokenRoomLinks == 0)
				{
					if (iMovingRoom != 0)
					{
						if (iMovingNewY != 1) { iMovingNewY--; }
							else { iMovingNewY = 24; }
						PlaySound ("wav/cross.wav");
					}
				} else {
					if (iChangingBrokenSide != 3)
					{
						iChangingBrokenSide = 3;
					} else {
						switch (iChangingBrokenRoom)
						{
							case 1: iChangingBrokenRoom = 21; break;
							case 2: iChangingBrokenRoom = 22; break;
							case 3: iChangingBrokenRoom = 23; break;
							case 4: iChangingBrokenRoom = 24; break;
							default: iChangingBrokenRoom -= 4; break;
						}
					}
				}
				break;
			case 3:
				iEventDoor = EventInfo (iChangeEvent, 2);
				if (iEventDoor > 10) { iEventDoor-=10; }
					else { iEventDoor+=20; }
				EventDoor (iEventDoor);
				break;
		}
	}

	if (strcmp (sAction, "down") == 0)
	{
		switch (iScreen)
		{
			case 1:
				if (iSelected <= 20) { iSelected+=10; }
					else { iSelected-=20; }
				break;
			case 2:
				if (iBrokenRoomLinks == 0)
				{
					if (iMovingRoom != 0)
					{
						if (iMovingNewY != 24) { iMovingNewY++; }
							else { iMovingNewY = 1; }
						PlaySound ("wav/cross.wav");
					}
				} else {
					if (iChangingBrokenSide != 4)
					{
						iChangingBrokenSide = 4;
					} else {
						switch (iChangingBrokenRoom)
						{
							case 21: iChangingBrokenRoom = 1; break;
							case 22: iChangingBrokenRoom = 2; break;
							case 23: iChangingBrokenRoom = 3; break;
							case 24: iChangingBrokenRoom = 4; break;
							default: iChangingBrokenRoom += 4; break;
						}
					}
				}
				break;
			case 3:
				iEventDoor = EventInfo (iChangeEvent, 2);
				if (iEventDoor <= 20) { iEventDoor+=10; }
					else { iEventDoor-=20; }
				EventDoor (iEventDoor);
				break;
		}
	}

	if (strcmp (sAction, "left bracket") == 0)
	{
		switch (iScreen)
		{
			case 2:
				if (iBrokenRoomLinks == 0)
				{
					iMovingNewBusy = 0;
					switch (iMovingRoom)
					{
						case 0: iMovingRoom = ROOMS; break;
						case 1: iMovingRoom = ROOMS; break;
						default: iMovingRoom--; break;
					}
				}
				break;
			case 3:
				iEventRoom = EventInfo (iChangeEvent, 1);
				if ((iEventRoom >= 2) && (iEventRoom <= 24))
				{
					EventRoom (iEventRoom - 1);
				} else {
					EventRoom (24);
				}
				break;
		}
	}

	if (strcmp (sAction, "right bracket") == 0)
	{
		switch (iScreen)
		{
			case 2:
				if (iBrokenRoomLinks == 0)
				{
					iMovingNewBusy = 0;
					switch (iMovingRoom)
					{
						case 0: iMovingRoom = 1; break;
						case 24: iMovingRoom = 1; break;
						default: iMovingRoom++; break;
					}
				}
				break;
			case 3:
				iEventRoom = EventInfo (iChangeEvent, 1);
				if ((iEventRoom >= 1) && (iEventRoom <= 23))
				{
					EventRoom (iEventRoom + 1);
				} else {
					EventRoom (1);
				}
				break;
		}
	}

	if (strcmp (sAction, "enter") == 0)
	{
		switch (iScreen)
		{
			case 1:
				ChangePos (iSelected);
				break;
			case 2:
				if (iBrokenRoomLinks == 0)
				{
					if (iMovingRoom != 0)
					{
						if (arRoomMoving[iMovingNewX][iMovingNewY] == 0)
						{
							RemoveOldRoom();
							AddNewRoom (iMovingNewX, iMovingNewY, iMovingRoom);
							iChanged++;
						}
						iMovingRoom = 0; iMovingNewBusy = 0;
					}
				} else {
					LinkPlus();
				}
				break;
		}
	}

	if (strcmp (sAction, "env") == 0)
	{
		/*** Nothing for now. ***/
	}
}
/*****************************************************************************/
void InitScreen (void)
/*****************************************************************************/
{
	SDL_AudioSpec fmt;
	char sImage[MAX_IMG + 2];
	SDL_Surface *imgicon;
	int iJoyNr;
	SDL_Rect barbox;
	SDL_Event event;
	DIR *dDir;
	struct dirent *stDirent;
	char sBlock[3 + 2]; /*** 2 chars, plus \0 ***/
	int iXPosOld, iYPosOld;
	int iLocX, iLocY;
	char sExtra[20 + 2];
	char sRoom[20 + 2];
	const Uint8 *keystate;
	Uint32 oldticks, newticks;

	/*** Used for looping. ***/
	int iLoopExtra;
	int iLoopTile;
	int iLoopRoom;
	int iLoopX, iLoopY;

	if (SDL_Init (SDL_INIT_AUDIO|SDL_INIT_VIDEO|
		SDL_INIT_GAMECONTROLLER|SDL_INIT_HAPTIC) < 0)
	{
		printf ("[FAILED] Unable to init SDL: %s!\n", SDL_GetError());
		exit (EXIT_ERROR);
	}
	atexit (SDL_Quit);

	/*** main window ***/
	window = SDL_CreateWindow (EDITOR_NAME " " EDITOR_VERSION,
		SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
		(WINDOW_WIDTH) * iScale, (WINDOW_HEIGHT) * iScale, iFullscreen);
	if (window == NULL)
	{
		printf ("[FAILED] Unable to create main window: %s!\n", SDL_GetError());
		exit (EXIT_ERROR);
	}
	iWindowID = SDL_GetWindowID (window);
	iActiveWindowID = iWindowID;
	ascreen = SDL_CreateRenderer (window, -1, 0);
	if (ascreen == NULL)
	{
		printf ("[FAILED] Unable to set video mode: %s!\n", SDL_GetError());
		exit (EXIT_ERROR);
	}

	/*** Some people may prefer linear, but we're going old school. ***/
	SDL_SetHint (SDL_HINT_RENDER_SCALE_QUALITY, "nearest");
	if (iFullscreen != 0)
	{
		SDL_RenderSetLogicalSize (ascreen, (WINDOW_WIDTH) * iScale,
			(WINDOW_HEIGHT) * iScale);
	}

	if (TTF_Init() == -1)
	{
		exit (EXIT_ERROR);
	}

	LoadFonts();

	curArrow = SDL_CreateSystemCursor (SDL_SYSTEM_CURSOR_ARROW);
	curWait = SDL_CreateSystemCursor (SDL_SYSTEM_CURSOR_WAIT);
	curHand = SDL_CreateSystemCursor (SDL_SYSTEM_CURSOR_HAND);

	if (iNoAudio != 1)
	{
		PrIfDe ("[  OK  ] Initializing Audio\n");
		fmt.freq = 44100;
		fmt.format = AUDIO_S16;
		fmt.channels = 2;
		fmt.samples = 512;
		fmt.callback = MixAudio;
		fmt.userdata = NULL;
		if (SDL_OpenAudio (&fmt, NULL) < 0)
		{
			printf ("[FAILED] Unable to open audio: %s!\n", SDL_GetError());
			exit (EXIT_ERROR);
		}
		SDL_PauseAudio (0);
	}

	/*** main window icon ***/
	snprintf (sImage, MAX_IMG, "%s%s", PNG_VARIOUS, "TODO_icon.png");
	imgicon = IMG_Load (sImage);
	if (imgicon == NULL)
	{
		printf ("[ WARN ] Could not load \"%s\": %s!\n", sImage, strerror (errno));
	} else {
		SDL_SetWindowIcon (window, imgicon);
	}

	/*** Open the first available controller. ***/
	iController = 0;
	if (iNoController != 1)
	{
		for (iJoyNr = 0; iJoyNr < SDL_NumJoysticks(); iJoyNr++)
		{
			if (SDL_IsGameController (iJoyNr))
			{
				controller = SDL_GameControllerOpen (iJoyNr);
				if (controller)
				{
					snprintf (sControllerName, MAX_CON, "%s",
						SDL_GameControllerName (controller));
					if (iDebug == 1)
					{
						printf ("[ INFO ] Found a controller \"%s\"; \"%s\".\n",
							sControllerName, SDL_GameControllerNameForIndex (iJoyNr));
					}
					joystick = SDL_GameControllerGetJoystick (controller);
					iController = 1;

					/*** Just for fun, use haptic. ***/
					if (SDL_JoystickIsHaptic (joystick))
					{
						haptic = SDL_HapticOpenFromJoystick (joystick);
						if (SDL_HapticRumbleInit (haptic) == 0)
						{
							SDL_HapticRumblePlay (haptic, 1.0, 1000);
						} else {
							printf ("[ WARN ] Could not initialize the haptic device: %s\n",
								SDL_GetError());
						}
					} else {
						PrIfDe ("[ INFO ] The game controller is not haptic.\n");
					}
				} else {
					printf ("[ WARN ] Could not open game controller %i: %s\n",
						iController, SDL_GetError());
				}
			}
		}
		if (iController != 1) { PrIfDe ("[ INFO ] No controller found.\n"); }
	} else {
		PrIfDe ("[ INFO ] Using keyboard and mouse.\n");
	}

	/*******************/
	/* Preload images. */
	/*******************/

	/*** loading ***/
	SDL_SetCursor (curWait);
	PreLoad (PNG_VARIOUS, "loading.png", &imgloading);
	PreLoad (PNG_VARIOUS, "hourglass_sprite.png", &imghourglasssprite);
	PreLoad (PNG_VARIOUS, "sand_sprite.png", &imgsandsprite);
	ShowImage (imgloading, 0, 0, "imgloading", ascreen, iScale, 1);
	SDL_SetRenderDrawColor (ascreen, 0x22, 0x22, 0x22, SDL_ALPHA_OPAQUE);
	barbox.x = 10 * iScale;
	barbox.y = 10 * iScale;
	barbox.w = 20 * iScale;
	barbox.h = 447 * iScale;
	SDL_RenderFillRect (ascreen, &barbox);
	iHourglassFrame = 1;
	iSandFrame = 1;
	SDL_RenderPresent (ascreen);

	iPreLoaded = 0;
	iCurrentBarHeight = 0;
	iNrToPreLoad = 76; /*** Value can be obtained via debug mode. ***/
	SDL_SetCursor (curWait);

	/*** buttons ***/
	PreLoad (PNG_BUTTONS, "left_0.png", &imgleft_0);
	PreLoad (PNG_BUTTONS, "left_1.png", &imgleft_1);
	PreLoad (PNG_BUTTONS, "right_0.png", &imgright_0);
	PreLoad (PNG_BUTTONS, "right_1.png", &imgright_1);
	PreLoad (PNG_BUTTONS, "up_0.png", &imgup_0);
	PreLoad (PNG_BUTTONS, "up_1.png", &imgup_1);
	PreLoad (PNG_BUTTONS, "down_0.png", &imgdown_0);
	PreLoad (PNG_BUTTONS, "down_1.png", &imgdown_1);
	PreLoad (PNG_BUTTONS, "left_right_no.png", &imglrno);
	PreLoad (PNG_BUTTONS, "up_down_no.png", &imgudno);
	if (iController == 0)
	{
		PreLoad (PNG_BUTTONS, "close_big_0.png", &imgclosebig_0);
		PreLoad (PNG_BUTTONS, "close_big_1.png", &imgclosebig_1);
		PreLoad (PNG_BUTTONS, "rooms_on_0.png", &imgroomson_0);
		PreLoad (PNG_BUTTONS, "rooms_on_1.png", &imgroomson_1);
		PreLoad (PNG_BUTTONS, "rooms_off.png", &imgroomsoff);
		PreLoad (PNG_BUTTONS, "broken_rooms_on_0.png", &imgbroomson_0);
		PreLoad (PNG_BUTTONS, "broken_rooms_on_1.png", &imgbroomson_1);
		PreLoad (PNG_BUTTONS, "broken_rooms_off.png", &imgbroomsoff);
		PreLoad (PNG_BUTTONS, "up_down_no_nfo.png", &imgudnonfo);
		PreLoad (PNG_BUTTONS, "events_on_0.png", &imgeventson_0);
		PreLoad (PNG_BUTTONS, "events_on_1.png", &imgeventson_1);
		PreLoad (PNG_BUTTONS, "events_off.png", &imgeventsoff);
		PreLoad (PNG_BUTTONS, "save_on_0.png", &imgsaveon_0);
		PreLoad (PNG_BUTTONS, "save_on_1.png", &imgsaveon_1);
		PreLoad (PNG_BUTTONS, "save_off.png", &imgsaveoff);
		PreLoad (PNG_BUTTONS, "quit_0.png", &imgquit_0);
		PreLoad (PNG_BUTTONS, "quit_1.png", &imgquit_1);
		PreLoad (PNG_BUTTONS, "prev_on_0.png", &imgprevon_0);
		PreLoad (PNG_BUTTONS, "prev_on_1.png", &imgprevon_1);
		PreLoad (PNG_BUTTONS, "prev_off.png", &imgprevoff);
		PreLoad (PNG_BUTTONS, "next_on_0.png", &imgnexton_0);
		PreLoad (PNG_BUTTONS, "next_on_1.png", &imgnexton_1);
		PreLoad (PNG_BUTTONS, "next_off.png", &imgnextoff);
	} else {
		PreLoad (PNG_GAMEPAD, "close_big_0.png", &imgclosebig_0);
		PreLoad (PNG_GAMEPAD, "close_big_1.png", &imgclosebig_1);
		PreLoad (PNG_GAMEPAD, "rooms_on_0.png", &imgroomson_0);
		PreLoad (PNG_GAMEPAD, "rooms_on_1.png", &imgroomson_1);
		PreLoad (PNG_GAMEPAD, "rooms_off.png", &imgroomsoff);
		PreLoad (PNG_GAMEPAD, "broken_rooms_on_0.png", &imgbroomson_0);
		PreLoad (PNG_GAMEPAD, "broken_rooms_on_1.png", &imgbroomson_1);
		PreLoad (PNG_GAMEPAD, "broken_rooms_off.png", &imgbroomsoff);
		PreLoad (PNG_GAMEPAD, "up_down_no_nfo.png", &imgudnonfo);
		PreLoad (PNG_GAMEPAD, "events_on_0.png", &imgeventson_0);
		PreLoad (PNG_GAMEPAD, "events_on_1.png", &imgeventson_1);
		PreLoad (PNG_GAMEPAD, "events_off.png", &imgeventsoff);
		PreLoad (PNG_GAMEPAD, "save_on_0.png", &imgsaveon_0);
		PreLoad (PNG_GAMEPAD, "save_on_1.png", &imgsaveon_1);
		PreLoad (PNG_GAMEPAD, "save_off.png", &imgsaveoff);
		PreLoad (PNG_GAMEPAD, "quit_0.png", &imgquit_0);
		PreLoad (PNG_GAMEPAD, "quit_1.png", &imgquit_1);
		PreLoad (PNG_GAMEPAD, "prev_on_0.png", &imgprevon_0);
		PreLoad (PNG_GAMEPAD, "prev_on_1.png", &imgprevon_1);
		PreLoad (PNG_GAMEPAD, "prev_off.png", &imgprevoff);
		PreLoad (PNG_GAMEPAD, "next_on_0.png", &imgnexton_0);
		PreLoad (PNG_GAMEPAD, "next_on_1.png", &imgnexton_1);
		PreLoad (PNG_GAMEPAD, "next_off.png", &imgnextoff);
	}

	/*** living ***/
	PreLoad (PNG_LIVING, "kid_r.png", &imgkidr[1]);
	PreLoad (PNG_LIVING, "kid_l.png", &imgkidl[1]);
	PreLoad (PNG_LIVING, "guard_l_bg.png", &imgguardbg[1]);
	PreLoad (PNG_LIVING, "guard_r_bg.png", &imgguardbg[2]);
	PreLoad (PNG_LIVING, "guard_l_pb.png", &imgguardpb[1]);
	PreLoad (PNG_LIVING, "guard_r_pb.png", &imgguardpb[2]);
	PreLoad (PNG_LIVING, "guard_l_unk.png", &imgguardunk[1]);
	PreLoad (PNG_LIVING, "guard_r_unk.png", &imgguardunk[2]);
	PreLoad (PNG_LIVING, "skeleton_l.png", &imgskel[1]);
	PreLoad (PNG_LIVING, "skeleton_r.png", &imgskel[2]);
	PreLoad (PNG_LIVING, "fat_l.png", &imgfat[1]);
	PreLoad (PNG_LIVING, "fat_r.png", &imgfat[2]);
	PreLoad (PNG_LIVING, "jaffar_l.png", &imgjaffar[1]);
	PreLoad (PNG_LIVING, "jaffar_r.png", &imgjaffar[2]);
	PreLoad (PNG_LIVING, "shadow_l.png", &imgshadow[1]);
	PreLoad (PNG_LIVING, "shadow_r.png", &imgshadow[2]);

	/*** sliving ***/
	PreLoad (PNG_SLIVING, "kid_r.png", &imgkidr[2]);
	PreLoad (PNG_SLIVING, "kid_l.png", &imgkidl[2]);
	PreLoad (PNG_SLIVING, "guard_l.png", &imgguardsel[1]);
	PreLoad (PNG_SLIVING, "guard_r.png", &imgguardsel[2]);
	PreLoad (PNG_SLIVING, "skeleton_l.png", &imgskelsel[1]);
	PreLoad (PNG_SLIVING, "skeleton_r.png", &imgskelsel[2]);
	PreLoad (PNG_SLIVING, "fat_l.png", &imgfatsel[1]);
	PreLoad (PNG_SLIVING, "fat_r.png", &imgfatsel[2]);
	PreLoad (PNG_SLIVING, "jaffar_l.png", &imgjaffarsel[1]);
	PreLoad (PNG_SLIVING, "jaffar_r.png", &imgjaffarsel[2]);
	PreLoad (PNG_SLIVING, "shadow_l.png", &imgshadowsel[1]);
	PreLoad (PNG_SLIVING, "shadow_r.png", &imgshadowsel[2]);

	/*** various ***/
	PreLoad (PNG_VARIOUS, "black.png", &imgblack);
	PreLoad (PNG_VARIOUS, "border_small_live.png", &imgbordersl);
	PreLoad (PNG_VARIOUS, "sel_room_current.png", &imgsrc);
	PreLoad (PNG_VARIOUS, "sel_room_start.png", &imgsrs);
	PreLoad (PNG_VARIOUS, "sel_room_moving.png", &imgsrm);
	PreLoad (PNG_VARIOUS, "sel_room_cross.png", &imgsrp);
	PreLoad (PNG_VARIOUS, "sel_room_broken.png", &imgsrb);
	PreLoad (PNG_VARIOUS, "sel_event.png", &imgsele);
	PreLoad (PNG_VARIOUS, "event_unused.png", &imgeventu);
	PreLoad (PNG_VARIOUS, "unknown.png", &imgunk);
	PreLoad (PNG_VARIOUS, "playtest.png", &imgplaytest);
	PreLoad (PNG_VARIOUS, "sel_block.png", &imgselblock);
	PreLoad (PNG_VARIOUS, "custom_hover.png", &imgchover);
	if (iController == 0)
	{
		PreLoad (PNG_VARIOUS, "l0003.png", &imgl0003t);
		PreLoad (PNG_VARIOUS, "l0406.png", &imgl0406t);
		PreLoad (PNG_VARIOUS, "l0709.png", &imgl0709t);
		PreLoad (PNG_VARIOUS, "l1011.png", &imgl1011t);
		PreLoad (PNG_VARIOUS, "l1213.png", &imgl1213t);
		PreLoad (PNG_VARIOUS, "l14.png", &imgl14t);
		PreLoad (PNG_VARIOUS, "l15.png", &imgl15t);
		PreLoad (PNG_VARIOUS, "level_bar.png", &imgbar);
		PreLoad (PNG_VARIOUS, "room_links.png", &imgrl);
		PreLoad (PNG_VARIOUS, "broken_room_links.png", &imgbrl);
		PreLoad (PNG_VARIOUS, "events.png", &imgevents);
	} else {
		PreLoad (PNG_GAMEPAD, "l0003.png", &imgl0003t);
		PreLoad (PNG_GAMEPAD, "l0406.png", &imgl0406t);
		PreLoad (PNG_GAMEPAD, "l0709.png", &imgl0709t);
		PreLoad (PNG_GAMEPAD, "l1011.png", &imgl1011t);
		PreLoad (PNG_GAMEPAD, "l1213.png", &imgl1213t);
		PreLoad (PNG_GAMEPAD, "l14.png", &imgl14t);
		PreLoad (PNG_GAMEPAD, "l15.png", &imgl15t);
		PreLoad (PNG_GAMEPAD, "level_bar.png", &imgbar);
		PreLoad (PNG_GAMEPAD, "room_links.png", &imgrl);
		PreLoad (PNG_GAMEPAD, "broken_room_links.png", &imgbrl);
		PreLoad (PNG_GAMEPAD, "events.png", &imgevents);
	}

	dDir = opendir (PNG_L0003);
	if (dDir == NULL)
	{
		printf ("[FAILED] Cannot open directory \"%s\": %s!\n",
			PNG_L0003, strerror (errno));
		exit (EXIT_ERROR);
	}
	while ((stDirent = readdir (dDir)) != NULL)
	{
		if ((strcmp (stDirent->d_name, ".") != 0) &&
			(strcmp (stDirent->d_name, "..") != 0))
		{
			snprintf (sBlock, 3, "%c%c",
				stDirent->d_name[0], stDirent->d_name[1]);
			PreLoad (PNG_L0003, stDirent->d_name,
				&imgl0003[(int)strtol(sBlock, NULL, 16)]);
		}
	}
	closedir (dDir);

	/*** extras ***/
	for (iLoopExtra = 0; iLoopExtra <= 10; iLoopExtra++)
	{
		snprintf (sExtra, 20, "extras_%02i.png", iLoopExtra);
		PreLoad (PNG_EXTRAS, sExtra, &extras[iLoopExtra]);
	}

	/*** 14x14 rooms ***/
	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		snprintf (sRoom, 20, "room%i.png", iLoopRoom);
		PreLoad (PNG_ROOMS, sRoom, &imgroom[iLoopRoom]);
	}

	/*** One final update of the bar. ***/
	LoadingBar (BAR_FULL);

	if (iDebug == 1)
		{ printf ("[ INFO ] Preloaded images: %i\n", iPreLoaded); }
	SDL_SetCursor (curArrow);

	iCurLevel = iStartLevel;
	if (iDebug == 1)
		{ printf ("[ INFO ] Starting in level: %i\n", iStartLevel); }

	iSelected = 1; /*** Start with the upper left selected. ***/
	iScreen = 1;
	iDownAt = 0;
	iChangeEvent = 1;
	oldticks = 0;

	ShowScreen (iScreen);
	InitPopUp();
	while (1)
	{
		if (iNoAnim == 0)
		{
			/*** This is for the animation; 12 fps (1000/83) is fine for PoP. ***/
			newticks = SDL_GetTicks();
			if (newticks > oldticks + 83)
			{
				/*** ... ***/
				ShowScreen (iScreen);
				oldticks = newticks;
			}
		}

		while (SDL_PollEvent (&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN: /*** https://wiki.libsdl.org/SDL_Keycode ***/
					switch (event.key.keysym.sym)
					{
						case SDLK_F1: if (iScreen == 1) { Help(); } break;
						case SDLK_F2: if (iScreen == 1) { EXE(); } break;
						case SDLK_KP_ENTER:
						case SDLK_RETURN:
						case SDLK_SPACE:
							if (((event.key.keysym.mod & KMOD_LALT) ||
								(event.key.keysym.mod & KMOD_RALT)) && (iScreen == 1))
							{
								Zoom (1);
								iExtras = 0;
								PlaySound ("wav/extras.wav");
							} else {
								InitScreenAction ("enter");
							}
							break;
						case SDLK_LEFT:
							if ((event.key.keysym.mod & KMOD_LSHIFT) ||
								(event.key.keysym.mod & KMOD_RSHIFT))
							{
								switch (iScreen)
								{
									case 1:
										if (arRoomLinks[iCurRoom][1] != 0)
										{
											iCurRoom = arRoomLinks[iCurRoom][1];
											PlaySound ("wav/scroll.wav");
										}
										break;
									case 3:
										ChangeEvent (-1, 0);
										break;
								}
							} else if ((event.key.keysym.mod & KMOD_LCTRL) ||
								(event.key.keysym.mod & KMOD_RCTRL))
							{
								if (iScreen == 3) { ChangeEvent (-10, 0); }
							} else {
								InitScreenAction ("left");
							}
							break;
						case SDLK_RIGHT:
							if ((event.key.keysym.mod & KMOD_LSHIFT) ||
								(event.key.keysym.mod & KMOD_RSHIFT))
							{
								switch (iScreen)
								{
									case 1:
										if (arRoomLinks[iCurRoom][2] != 0)
										{
											iCurRoom = arRoomLinks[iCurRoom][2];
											PlaySound ("wav/scroll.wav");
										}
										break;
									case 3:
										ChangeEvent (1, 0);
										break;
								}
							} else if ((event.key.keysym.mod & KMOD_LCTRL) ||
								(event.key.keysym.mod & KMOD_RCTRL))
							{
								if (iScreen == 3) { ChangeEvent (10, 0); }
							} else {
								InitScreenAction ("right");
							}
							break;
						case SDLK_UP:
							if ((event.key.keysym.mod & KMOD_LSHIFT) ||
								(event.key.keysym.mod & KMOD_RSHIFT))
							{
								if (iScreen == 1)
								{
									if (arRoomLinks[iCurRoom][3] != 0)
									{
										iCurRoom = arRoomLinks[iCurRoom][3];
										PlaySound ("wav/scroll.wav");
									}
								}
							} else {
								InitScreenAction ("up");
							}
							break;
						case SDLK_DOWN:
							if ((event.key.keysym.mod & KMOD_LSHIFT) ||
								(event.key.keysym.mod & KMOD_RSHIFT))
							{
								if (iScreen == 1)
								{
									if (arRoomLinks[iCurRoom][4] != 0)
									{
										iCurRoom = arRoomLinks[iCurRoom][4];
										PlaySound ("wav/scroll.wav");
									}
								}
							} else {
								InitScreenAction ("down");
							}
							break;
						case SDLK_LEFTBRACKET:
							InitScreenAction ("left bracket");
							break;
						case SDLK_RIGHTBRACKET:
							InitScreenAction ("right bracket");
							break;
						case SDLK_MINUS:
						case SDLK_KP_MINUS:
							if ((iChanged != 0) && (iCurLevel != 1)) { InitPopUpSave(); }
							Prev(); break;
						case SDLK_KP_PLUS:
						case SDLK_EQUALS:
							if ((iChanged != 0) && (iCurLevel != 0)) { InitPopUpSave(); }
							Next(); break;
						case SDLK_QUOTE:
							if (iScreen == 1)
							{
								if ((event.key.keysym.mod & KMOD_LSHIFT) ||
									(event.key.keysym.mod & KMOD_RSHIFT))
								{
									Sprinkle();
									PlaySound ("wav/extras.wav");
									iChanged++;
								} else {
									SetLocation (iCurRoom, iSelected, iLastBlock, iLastModif);
									PlaySound ("wav/ok_close.wav");
									iChanged++;
								}
							}
							break;
						case SDLK_d: RunLevel (iCurLevel); break;
						case SDLK_s:
							SaveLevel (iCurLevel);
							break;
						case SDLK_t:
							if (iScreen == 1) { InitScreenAction ("env"); }
							break;
						case SDLK_r:
							if (iScreen != 2)
							{
								iScreen = 2;
								iMovingRoom = 0;
								iMovingNewBusy = 0;
								iChangingBrokenRoom = iCurRoom;
								iChangingBrokenSide = 1;
								PlaySound ("wav/screen2or3.wav");
							} else if (iBrokenRoomLinks == 0) {
								iBrokenRoomLinks = 1;
								PlaySound ("wav/screen2or3.wav");
							}
							break;
						case SDLK_e:
							if (iScreen == 2)
							{
								iBrokenRoomLinks = BrokenRoomLinks (0); /*** Why? ***/
							}
							if (iScreen != 3)
							{
								iScreen = 3;
								PlaySound ("wav/screen2or3.wav");
							}
							break;
						case SDLK_c:
							if (iScreen == 1)
							{
								if ((event.key.keysym.mod & KMOD_LCTRL) ||
									(event.key.keysym.mod & KMOD_RCTRL))
								{
									CopyPaste (iCurRoom, 1);
									PlaySound ("wav/extras.wav");
								}
							}
							break;
						case SDLK_z:
							if (iScreen == 1)
							{
								Zoom (0);
								iExtras = 0;
								PlaySound ("wav/extras.wav");
							}
							break;
						case SDLK_f:
							if (iScreen == 1)
							{
								Zoom (1);
								iExtras = 0;
								PlaySound ("wav/extras.wav");
							}
							break;
						case SDLK_h:
							if (iScreen == 1)
							{
								FlipRoom (iCurRoom, 1);
								PlaySound ("wav/extras.wav");
								iChanged++;
							}
							break;
						case SDLK_v:
							if (iScreen == 1)
							{
								if ((event.key.keysym.mod & KMOD_LCTRL) ||
									(event.key.keysym.mod & KMOD_RCTRL))
								{
									CopyPaste (iCurRoom, 2);
									PlaySound ("wav/extras.wav");
									iChanged++;
								} else {
									FlipRoom (iCurRoom, 2);
									PlaySound ("wav/extras.wav");
									iChanged++;
								}
							}
							break;
						case SDLK_y:
							if (iScreen == 3)
							{
								if (EventInfo (iChangeEvent, 3) == 0) { EventNext (1); }
							}
							break;
						case SDLK_n:
							if (iScreen == 3)
							{
								if (EventInfo (iChangeEvent, 3) == 1) { EventNext (0); }
							}
							break;
						case SDLK_BACKSPACE:
							if ((iScreen == 2) && (iBrokenRoomLinks == 1))
								{ LinkMinus(); }
							break;
						case SDLK_ESCAPE:
						case SDLK_q:
							switch (iScreen)
							{
								case 1:
									Quit(); break;
								case 2:
									/*** TODO: iBrokenRoomLinks = BrokenRoomLinks (0); ***/
									/*** no break ***/
								case 3:
									iScreen = 1; break;
							}
							break;
						case SDLK_BACKSLASH:
							if (iScreen == 1)
							{
								/*** Randomize the entire level. ***/
								for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
								{
									for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
									{
										UseTile (-1, iLoopTile, iLoopRoom);
									}
								}
								PlaySound ("wav/ok_close.wav");
								iChanged++;
							}
							break;
						case SDLK_SLASH:
							if (iScreen == 1) { ClearRoom(); }
							break;
					}
					ShowScreen (iScreen);
					break;
				case SDL_MOUSEMOTION:
					iXPosOld = iXPos;
					iYPosOld = iYPos;
					iXPos = event.motion.x;
					iYPos = event.motion.y;
					if ((iXPosOld == iXPos) && (iYPosOld == iYPos)) { break; }

					if (OnLevelBar() == 1)
						{ iPlaytest = 1; } else { iPlaytest = 0; }

					if (iScreen == 1)
					{
						/*** hover ***/
						for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
						{
							if ((iLoopTile >= 1) && (iLoopTile <= 10))
								{ iLocX = iLoopTile; iLocY = 1; }
							if ((iLoopTile >= 11) && (iLoopTile <= 20))
								{ iLocX = iLoopTile - 10; iLocY = 2; }
							if ((iLoopTile >= 21) && (iLoopTile <= 30))
								{ iLocX = iLoopTile - 20; iLocY = 3; }

							if ((InArea (25 + (64 * (iLocX - 1)), 58 + (128 * (iLocY - 1)),
								25 + (64 * iLocX), 58 + (128 * iLocY)) == 1) &&
								(iSelected != iLoopTile)) { iSelected = iLoopTile; }
						}

						/*** extras ***/
						if (InArea (608, 3, 608 + 49, 3 + 19) == 0) { iExtras = 0; }
						if (InArea (608, 3, 608 + 9, 3 + 9) == 1) { iExtras = 1; }
						if (InArea (618, 3, 618 + 9, 3 + 9) == 1) { iExtras = 2; }
						if (InArea (628, 3, 628 + 9, 3 + 9) == 1) { iExtras = 3; }
						if (InArea (638, 3, 638 + 9, 3 + 9) == 1) { iExtras = 4; }
						if (InArea (648, 3, 648 + 9, 3 + 9) == 1) { iExtras = 5; }
						if (InArea (608, 13, 608 + 9, 13 + 9) == 1) { iExtras = 6; }
						if (InArea (618, 13, 618 + 9, 13 + 9) == 1) { iExtras = 7; }
						if (InArea (628, 13, 628 + 9, 13 + 9) == 1) { iExtras = 8; }
						if (InArea (638, 13, 638 + 9, 13 + 9) == 1) { iExtras = 9; }
						if (InArea (648, 13, 648 + 9, 13 + 9) == 1) { iExtras = 10; }
					}

					if (iScreen == 2)
					{
						/*** if (iMovingRoom != 0) { ShowScreen (2); } ***/
					}

					ShowScreen (iScreen);
					break;
				case SDL_MOUSEBUTTONDOWN:
					if (event.button.button == 1)
					{
						if (InArea (0, 50, 0 + 25, 50 + 392) == 1) /*** left ***/
						{
							if (arRoomLinks[iCurRoom][1] != 0) { iDownAt = 1; }
						}
						if (InArea (665, 50, 665 + 25, 50 + 392) == 1) /*** right ***/
						{
							if (arRoomLinks[iCurRoom][2] != 0) { iDownAt = 2; }
						}
						if (InArea (25, 25, 25 + 640, 25 + 25) == 1) /*** up ***/
						{
							if (arRoomLinks[iCurRoom][3] != 0) { iDownAt = 3; }
						}
						if (InArea (25, 442, 25 + 640, 442 + 25) == 1) /*** down ***/
						{
							if (arRoomLinks[iCurRoom][4] != 0) { iDownAt = 4; }
						}
						if (InArea (0, 25, 0 + 25, 25 + 25) == 1) /*** rooms ***/
						{
							if (iBrokenRoomLinks == 0)
								{ iDownAt = 5; } else { iDownAt = 11; }
						}
						if (InArea (665, 25, 665 + 25, 25 + 25) == 1) /*** events ***/
							{ iDownAt = 6; }
						if (InArea (0, 442, 0 + 25, 442 + 25) == 1) /*** save ***/
							{ iDownAt = 7; }
						if (InArea (665, 442, 665 + 25, 442 + 25) == 1) /*** quit ***/
							{ iDownAt = 8; }
						if (InArea (0, 0, 0 + 25, 0 + 25) == 1) /*** prev level ***/
							{ iDownAt = 9; }
						if (InArea (665, 0, 665 + 25, 0 + 25) == 1) /*** next level ***/
							{ iDownAt = 10; }

						if (iScreen == 2)
						{
							if (iBrokenRoomLinks == 0)
							{
								for (iLoopX = 1; iLoopX <= ROOMS; iLoopX++)
								{
									for (iLoopY = 1; iLoopY <= ROOMS; iLoopY++)
									{
										if (InArea (290 + ((iLoopX - 1) * 15),
											66 + ((iLoopY - 1) * 15),
											290 + ((iLoopX - 1) * 15) + 14,
											66 + ((iLoopY - 1) * 15) + 14) == 1)
										{
											if (arRoomMoving[iLoopX][iLoopY] != 0)
											{
												iMovingNewBusy = 0;
												iMovingRoom = arRoomMoving[iLoopX][iLoopY];
											}
										}
									}
								}

								/*** side pane ***/
								for (iLoopY = 1; iLoopY <= ROOMS; iLoopY++)
								{
									if (InArea (267, 66 + ((iLoopY - 1) * 15),
										267 + 14, 66 + ((iLoopY - 1) * 15) + 14) == 1)
									{
										if (arRoomMoving[25][iLoopY] != 0)
										{
											iMovingNewBusy = 0;
											iMovingRoom = arRoomMoving[25][iLoopY];
										}
									}
								}

								if (InArea (622, 68, 622 + 25, 68 + 25) == 1)
									{ iDownAt = 11; } /*** rooms broken ***/
							} else {
								MouseSelectAdj();
							}
						}
					}
					ShowScreen (iScreen);
					break;
				case SDL_MOUSEBUTTONUP:
					iDownAt = 0;
					if (event.button.button == 1)
					{
						if (InArea (0, 50, 0 + 25, 50 + 392) == 1) /*** left ***/
						{
							if (arRoomLinks[iCurRoom][1] != 0)
							{
								iCurRoom = arRoomLinks[iCurRoom][1];
								PlaySound ("wav/scroll.wav");
							}
						}
						if (InArea (665, 50, 665 + 25, 50 + 392) == 1) /*** right ***/
						{
							if (arRoomLinks[iCurRoom][2] != 0)
							{
								iCurRoom = arRoomLinks[iCurRoom][2];
								PlaySound ("wav/scroll.wav");
							}
						}
						if (InArea (25, 25, 25 + 640, 25 + 25) == 1) /*** up ***/
						{
							if (arRoomLinks[iCurRoom][3] != 0)
							{
								iCurRoom = arRoomLinks[iCurRoom][3];
								PlaySound ("wav/scroll.wav");
							}
						}
						if (InArea (25, 442, 25 + 640, 442 + 25) == 1) /*** down ***/
						{
							if (arRoomLinks[iCurRoom][4] != 0)
							{
								iCurRoom = arRoomLinks[iCurRoom][4];
								PlaySound ("wav/scroll.wav");
							}
						}
						if (InArea (0, 25, 0 + 25, 25 + 25) == 1) /*** rooms ***/
						{
							if (iScreen != 2)
							{
								iScreen = 2; iMovingRoom = 0; iMovingNewBusy = 0;
								iChangingBrokenRoom = iCurRoom;
								iChangingBrokenSide = 1;
								PlaySound ("wav/screen2or3.wav");
							}
						}
						if (InArea (665, 25, 665 + 25, 25 + 25) == 1) /*** events ***/
						{
							if (iScreen == 2)
							{
								iBrokenRoomLinks = BrokenRoomLinks (0);
							}
							if (iScreen != 3)
							{
								iScreen = 3;
								PlaySound ("wav/screen2or3.wav");
							}
						}
						if (InArea (0, 442, 0 + 25, 442 + 25) == 1) /*** save ***/
						{
							SaveLevel (iCurLevel);
						}
						if (InArea (665, 442, 665 + 25, 442 + 25) == 1) /*** quit ***/
						{
							switch (iScreen)
							{
								case 1:
									Quit(); break;
								case 2:
									iBrokenRoomLinks = BrokenRoomLinks (0);
									iScreen = 1; break;
								case 3:
									iScreen = 1; break;
							}
						}
						if (InArea (0, 0, 0 + 25, 0 + 25) == 1) /*** prev level ***/
						{
							if ((iChanged != 0) && (iCurLevel != 1)) { InitPopUpSave(); }
							Prev();
							break; /*** Stop processing SDL_MOUSEBUTTONUP. ***/
						}
						if (InArea (665, 0, 665 + 25, 0 + 25) == 1) /*** next level ***/
						{
							if ((iChanged != 0) && (iCurLevel != 0)) { InitPopUpSave(); }
							Next();
							break; /*** Stop processing SDL_MOUSEBUTTONUP. ***/
						}
						if (OnLevelBar() == 1)
						{
							RunLevel (iCurLevel);
							break; /*** Stop processing SDL_MOUSEBUTTONUP. ***/
						}

						if (iScreen == 1)
						{
							if (InArea (25, 50, 25 + 640, 50 + 392) == 1) /*** tiles ***/
							{
								keystate = SDL_GetKeyboardState (NULL);
								if ((keystate[SDL_SCANCODE_LSHIFT]) ||
									(keystate[SDL_SCANCODE_RSHIFT]))
								{
									SetLocation (iCurRoom, iSelected, iLastBlock, iLastModif);
									PlaySound ("wav/ok_close.wav");
									iChanged++;
								} else {
									ChangePos (iSelected);
								}
							}

							/*** 1 ***/
							if (InArea (608, 3, 608 + 9, 3 + 9) == 1)
							{
								Zoom (0);
								iExtras = 0;
								PlaySound ("wav/extras.wav");
							}

							/*** 2 ***/
							if (InArea (618, 3, 618 + 9, 3 + 9) == 1)
							{
								CopyPaste (iCurRoom, 1);
								PlaySound ("wav/extras.wav");
							}

							/*** 3 ***/
							if (InArea (628, 3, 628 + 9, 3 + 9) == 1)
							{
								FlipRoom (iCurRoom, 2);
								PlaySound ("wav/extras.wav");
								iChanged++;
							}

							/*** 4 ***/
							if (InArea (638, 3, 638 + 9, 3 + 9) == 1)
								{ InitScreenAction ("env"); }

							/*** 5 ***/
							if (InArea (648, 3, 648 + 9, 3 + 9) == 1) { Help(); }

							/*** 6 ***/
							if (InArea (608, 13, 608 + 9, 13 + 9) == 1)
							{
								Sprinkle();
								PlaySound ("wav/extras.wav");
								iChanged++;
							}

							/*** 7 ***/
							if (InArea (618, 13, 618 + 9, 13 + 9) == 1)
							{
								CopyPaste (iCurRoom, 2);
								PlaySound ("wav/extras.wav");
								iChanged++;
							}

							/*** 8 ***/
							if (InArea (628, 13, 628 + 9, 13 + 9) == 1)
							{
								FlipRoom (iCurRoom, 1);
								PlaySound ("wav/extras.wav");
								iChanged++;
							}

							/*** 9 ***/
							if (InArea (638, 13, 638 + 9, 13 + 9) == 1)
								{ /*** Undo. ***/ }

							/*** 10 ***/
							if (InArea (648, 13, 648 + 9, 13 + 9) == 1) { EXE(); }
						}

						if (iScreen == 2)
						{
							if (iBrokenRoomLinks == 0)
							{
								for (iLoopX = 1; iLoopX <= ROOMS; iLoopX++)
								{
									for (iLoopY = 1; iLoopY <= ROOMS; iLoopY++)
									{
										if (InArea (290 + ((iLoopX - 1) * 15),
											66 + ((iLoopY - 1) * 15),
											290 + ((iLoopX - 1) * 15) + 14,
											66 + ((iLoopY - 1) * 15) + 14) == 1)
										{
											if (iMovingRoom != 0)
											{
												if (arRoomMoving[iLoopX][iLoopY] == 0)
												{
													RemoveOldRoom();
													AddNewRoom (iLoopX, iLoopY, iMovingRoom);
													iChanged++;
												}
												iMovingRoom = 0; iMovingNewBusy = 0;
											}
										}
									}
								}

								/*** side pane ***/
								for (iLoopY = 1; iLoopY <= ROOMS; iLoopY++)
								{
									if (InArea (267, 66 + ((iLoopY - 1) * 15),
										267 + 14, 66 + ((iLoopY - 1) * 15) + 14) == 1)
									{
										if (iMovingRoom != 0)
										{
											if (arRoomMoving[25][iLoopY] == 0)
											{
												RemoveOldRoom();
												AddNewRoom (25, iLoopY, iMovingRoom);
												iChanged++;
											}
											iMovingRoom = 0; iMovingNewBusy = 0;
										}
									}
								}

								if (InArea (622, 68, 622 + 25, 68 + 25) == 1)
								{ /*** rooms broken ***/
									iBrokenRoomLinks = 1;
									PlaySound ("wav/screen2or3.wav");
								}
							} else {
								if (MouseSelectAdj() == 1) { LinkPlus(); }
							}
						}

						if (iScreen == 3)
						{
							/*** event number ***/
							if (InArea (336, 65, 336 + 13, 65 + 20) == 1)
								{ ChangeEvent (-10, 0); }
							if (InArea (351, 65, 351 + 13, 65 + 20) == 1)
								{ ChangeEvent (-1, 0); }
							if (InArea (421, 65, 421 + 13, 65 + 20) == 1)
								{ ChangeEvent (1, 0); }
							if (InArea (436, 65, 436 + 13, 65 + 20) == 1)
								{ ChangeEvent (10, 0); }

							/*** room ***/
							for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
							{
								if (InArea (275 + (15 * iLoopRoom), 114,
									275 + (15 * iLoopRoom) + 14, 114 + 14) == 1)
									{ EventRoom (iLoopRoom); }
							}

							/*** tile ***/
							for (iLoopTile = 1; iLoopTile <= 10; iLoopTile++)
							{
								if (InArea (440 + (15 * iLoopTile), 160,
									440 + (15 * iLoopTile) + 14, 160 + 14) == 1)
									{ EventDoor (iLoopTile); }
							}
							for (iLoopTile = 11; iLoopTile <= 20; iLoopTile++)
							{
								if (InArea (440 + (15 * (iLoopTile - 10)), 175,
									440 + (15 * (iLoopTile - 10)) + 14, 175 + 14) == 1)
									{ EventDoor (iLoopTile); }
							}
							for (iLoopTile = 21; iLoopTile <= 30; iLoopTile++)
							{
								if (InArea (440 + (15 * (iLoopTile - 20)), 190,
									440 + (15 * (iLoopTile - 20)) + 14, 190 + 14) == 1)
									{ EventDoor (iLoopTile); }
							}

							/*** next ***/
							if (InArea (575, 230, 575 + 14, 230 + 14) == 1)
								{ EventNext (0); } /*** n ***/
							if (InArea (590, 230, 590 + 14, 230 + 14) == 1)
								{ EventNext (1); } /*** y ***/
						}
					}
					if (event.button.button == 2)
					{
						if (iScreen == 1) { ClearRoom(); }
					}
					if (event.button.button == 3)
					{
						if (iScreen == 1)
						{
							/*** Randomize the entire level. ***/
							for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
							{
								for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
								{
									UseTile (-1, iLoopTile, iLoopRoom);
								}
							}
							PlaySound ("wav/ok_close.wav");
							iChanged++;
						}
						if (iScreen == 2)
						{
							if (iBrokenRoomLinks == 1)
							{
								if (MouseSelectAdj() == 1) { LinkMinus(); }
							}
						}
					}
					ShowScreen (iScreen);
					break;
				case SDL_MOUSEWHEEL:
					if (event.wheel.y > 0) /*** scroll wheel up ***/
					{
						if (InArea (25, 50, 25 + 640, 50 + 392) == 1) /*** tiles ***/
						{
							keystate = SDL_GetKeyboardState (NULL);
							if ((keystate[SDL_SCANCODE_LSHIFT]) ||
								(keystate[SDL_SCANCODE_RSHIFT]))
							{ /*** right ***/
								if (arRoomLinks[iCurRoom][2] != 0)
								{
									iCurRoom = arRoomLinks[iCurRoom][2];
									PlaySound ("wav/scroll.wav");
								}
							} else { /*** up ***/
								if (arRoomLinks[iCurRoom][3] != 0)
								{
									iCurRoom = arRoomLinks[iCurRoom][3];
									PlaySound ("wav/scroll.wav");
								}
							}
						}
					}
					if (event.wheel.y < 0) /*** scroll wheel down ***/
					{
						if (InArea (25, 50, 25 + 640, 50 + 392) == 1) /*** tiles ***/
						{
							keystate = SDL_GetKeyboardState (NULL);
							if ((keystate[SDL_SCANCODE_LSHIFT]) ||
								(keystate[SDL_SCANCODE_RSHIFT]))
							{ /*** left ***/
								if (arRoomLinks[iCurRoom][1] != 0)
								{
									iCurRoom = arRoomLinks[iCurRoom][1];
									PlaySound ("wav/scroll.wav");
								}
							} else { /*** down ***/
								if (arRoomLinks[iCurRoom][4] != 0)
								{
									iCurRoom = arRoomLinks[iCurRoom][4];
									PlaySound ("wav/scroll.wav");
								}
							}
						}
					}
					ShowScreen (iScreen);
					break;
				case SDL_WINDOWEVENT:
					switch (event.window.event)
					{
						case SDL_WINDOWEVENT_EXPOSED:
							ShowScreen (iScreen); break;
						case SDL_WINDOWEVENT_CLOSE:
							Quit(); break;
						case SDL_WINDOWEVENT_FOCUS_GAINED:
							iActiveWindowID = iWindowID; break;
					}
					break;
				case SDL_QUIT:
					Quit(); break;
			}
		}

		/*** prevent CPU eating ***/
		gamespeed = REFRESH;
		while ((SDL_GetTicks() - looptime) < gamespeed)
		{
			SDL_Delay (10);
		}
		looptime = SDL_GetTicks();
	}
}
/*****************************************************************************/
void Quit (void)
/*****************************************************************************/
{
	if (iChanged != 0) { InitPopUpSave(); }
	TTF_CloseFont (font11);
	TTF_CloseFont (font15);
	TTF_CloseFont (font20);
	TTF_Quit();
	SDL_Quit();
	exit (EXIT_NORMAL);
}
/*****************************************************************************/
void GetBlockName (int iBlock, int iModif, char *sName)
/*****************************************************************************/
{
	char sBlock[5 + 2];
	char sModif[2 + 2];

	switch (iBlock)
	{
		case 0x00: snprintf (sBlock, 7, "%s", "WallM"); break;
		case 0x02: snprintf (sBlock, 7, "%s", "WallL"); break;
		case 0x05: snprintf (sBlock, 7, "%s", "Flr05"); break;
		case 0x07: snprintf (sBlock, 7, "%s", "Flr07"); break;
		case 0x2D: snprintf (sBlock, 7, "%s", "Flr2D"); break;
		case 0x35: snprintf (sBlock, 7, "%s", "Emp35"); break;
		case 0x39: snprintf (sBlock, 7, "%s", "Emp39"); break;
		case 0x3A: snprintf (sBlock, 7, "%s", "GateR"); break;
		default: snprintf (sBlock, 7, "%05X", iBlock); break;
	}
	snprintf (sModif, 4, "%02i", iModif);
	snprintf (sName, 9, "%s%s", sBlock, sModif);
}
/*****************************************************************************/
void GetFlagName (int iFlag, char *sName)
/*****************************************************************************/
{
	switch (iFlag)
	{
		case 0x00: snprintf (sName, 7, "%s", "empty"); break;
		case 0x01: snprintf (sName, 7, "%s", "floor"); break;
		case 0x02: snprintf (sName, 7, "%s", "wall "); break;
		case 0x03: snprintf (sName, 7, "%s", "fl+wl"); break;
		case 0x04: snprintf (sName, 7, "%s", "torch"); break;
		case 0x05: snprintf (sName, 7, "%s", "fl+tr"); break;
		case 0x06: snprintf (sName, 7, "%s", "wl+tr"); break;
		case 0x07: snprintf (sName, 7, "%s", "all  "); break;
		default: snprintf (sName, 7, "%05X", iFlag); break;
	}
}
/*****************************************************************************/
void GetObjectName (int iObject, char *sName)
/*****************************************************************************/
{
	switch (iObject)
	{
		case 0x00: snprintf (sName, 7, "%s", ""); break;
		case 0x08: snprintf (sName, 7, "%s", "loose"); break;
		case 0x10: snprintf (sName, 7, "%s", "raise"); break;
		case 0x18: snprintf (sName, 7, "%s", "drop "); break;
		case 0x20: snprintf (sName, 7, "%s", "gate "); break;
		case 0x28: snprintf (sName, 7, "%s", "spike"); break;
		case 0x30: snprintf (sName, 7, "%s", "heal "); break;
		case 0x38: snprintf (sName, 7, "%s", "hurt "); break;
		case 0x40: snprintf (sName, 7, "%s", "life "); break;
		case 0x48: snprintf (sName, 7, "%s", "flip "); break;
		case 0x50: snprintf (sName, 7, "%s", "float"); break;
		case 0x58: snprintf (sName, 7, "%s", "chomp"); break;
		case 0x60: snprintf (sName, 7, "%s", "sword"); break;
		case 0x68: snprintf (sName, 7, "%s", "mirr "); break;
		case 0x70: snprintf (sName, 7, "%s", "lexit"); break;
		case 0x78: snprintf (sName, 7, "%s", "skel "); break;
		case 0x80: snprintf (sName, 7, "%s", "TODO "); break;
		case 0x88: snprintf (sName, 7, "%s", "stars"); break;
		default: snprintf (sName, 7, "%05X", iObject); break;
	}
}
/*****************************************************************************/
char cShowDirection (int iDirection)
/*****************************************************************************/
{
	switch (iDirection)
	{
		case 0x00: return ('r'); break;
		case 0xFF: return ('l'); break;
		default: return ('?'); break;
	}
}
/*****************************************************************************/
void LoadFonts (void)
/*****************************************************************************/
{
	font11 = TTF_OpenFont ("ttf/Bitstream-Vera-Sans-Bold.ttf", 11 * iScale);
	font15 = TTF_OpenFont ("ttf/Bitstream-Vera-Sans-Bold.ttf", 15 * iScale);
	font20 = TTF_OpenFont ("ttf/Bitstream-Vera-Sans-Bold.ttf", 20 * iScale);
	if ((font11 == NULL) || (font15 == NULL) || (font20 == NULL))
	{
		printf ("[FAILED] TTF_OpenFont() failed!\n");
		exit (EXIT_ERROR);
	}
}
/*****************************************************************************/
void MixAudio (void *unused, Uint8 *stream, int iLen)
/*****************************************************************************/
{
	int iTemp;
	int iAmount;

	if (unused != NULL) { } /*** To prevent warnings. ***/

	SDL_memset (stream, 0, iLen); /*** SDL2 ***/
	for (iTemp = 0; iTemp < NUM_SOUNDS; iTemp++)
	{
		iAmount = (sounds[iTemp].dlen-sounds[iTemp].dpos);
		if (iAmount > iLen)
		{
			iAmount = iLen;
		}
		SDL_MixAudio (stream, &sounds[iTemp].data[sounds[iTemp].dpos], iAmount,
			SDL_MIX_MAXVOLUME);
		sounds[iTemp].dpos += iAmount;
	}
}
/*****************************************************************************/
void PreLoad (char *sPath, char *sPNG, SDL_Texture **imgImage)
/*****************************************************************************/
{
	char sImage[MAX_IMG + 2];
	int iBarHeight;

	snprintf (sImage, MAX_IMG, "%s%s", sPath, sPNG);
	*imgImage = IMG_LoadTexture (ascreen, sImage);
	if (!*imgImage)
	{
		printf ("[FAILED] IMG_LoadTexture: %s!\n", IMG_GetError());
		exit (EXIT_ERROR);
	}

	iPreLoaded++;
	iBarHeight = (int)(((float)iPreLoaded/(float)iNrToPreLoad) * BAR_FULL);
	if (iBarHeight >= iCurrentBarHeight + 10) { LoadingBar (iBarHeight); }
}
/*****************************************************************************/
void ShowImage (SDL_Texture *img, int iX, int iY, char *sImageInfo,
	SDL_Renderer *screen, float fMultiply, int iXYScale)
/*****************************************************************************/
{
	/* Usually, fMultiply is either iScale or ZoomGet().
	 *
	 * Also, usually, iXYScale is 1 with iScale and 0 with ZoomGet().
	 *
	 * If you use this function, make sure to verify the image is properly
	 * (re)scaled and (re)positioned when the main window is zoomed ("z")!
	 */

	SDL_Rect dest;
	SDL_Rect loc;
	int iWidth, iHeight;

	SDL_QueryTexture (img, NULL, NULL, &iWidth, &iHeight);
	loc.x = 0; loc.y = 0; loc.w = iWidth; loc.h = iHeight;
	if (iXYScale == 0)
	{
		dest.x = iX;
		dest.y = iY;
	} else {
		dest.x = iX * fMultiply;
		dest.y = iY * fMultiply;
	}
	dest.w = iWidth * fMultiply;
	dest.h = iHeight * fMultiply;

	/*** This is for the animation. ***/
	if (strcmp (sImageInfo, "imghourglasssprite") == 0)
	{
		loc.x = (iHourglassFrame - 1) * 38;
		loc.w = loc.w / 7;
		dest.w = dest.w / 7;
	}
	if (strcmp (sImageInfo, "imgsandsprite") == 0)
	{
		loc.x = (iSandFrame - 1) * 5;
		loc.w = loc.w / 3;
		dest.w = dest.w / 3;
	}
	if (strcmp (sImageInfo, "imgstatusbarsprite") == 0)
	{
		loc.x = (iStatusBarFrame - 1) * 20;
		loc.w = loc.w / 18;
		dest.w = dest.w / 18;
	}

	if (SDL_RenderCopy (screen, img, &loc, &dest) != 0)
	{
		printf ("[ WARN ] SDL_RenderCopy (%s): %s\n", sImageInfo, SDL_GetError());
	}
}
/*****************************************************************************/
void LoadingBar (int iBarHeight)
/*****************************************************************************/
{
	SDL_Rect bar;

	/*** bar ***/
	bar.x = 12 * iScale;
	bar.y = (455 - iBarHeight) * iScale;
	bar.w = 16 * iScale;
	bar.h = iBarHeight * iScale;
	SDL_SetRenderDrawColor (ascreen, 0x44, 0x44, 0x44, SDL_ALPHA_OPAQUE);
	SDL_RenderFillRect (ascreen, &bar);
	iCurrentBarHeight = iBarHeight;

	/*** hourglass sprite ***/
	iHourglassFrame = floor (((float)iBarHeight/BAR_FULL) * 7);
	ShowImage (imghourglasssprite, 326, 178, "imghourglasssprite",
		ascreen, iScale, 1);

	/*** sand sprite ***/
	if (iBarHeight != BAR_FULL)
	{
		ShowImage (imgsandsprite, 342, 206, "imgsandsprite",
		ascreen, iScale, 1);
		iSandFrame++;
		if (iSandFrame == 4) { iSandFrame = 1; }
	}

	/*** refresh screen ***/
	SDL_RenderPresent (ascreen);
}
/*****************************************************************************/
void ShowScreen (int iScreenS)
/*****************************************************************************/
{
	char sLevel[MAX_TEXT + 2];
	char arText[1 + 2][MAX_TEXT + 2];
	int iImgDir, iImgX, iImgY;
	int iUnusedRooms;
	int iXShow, iYShow;
	char sShowRoom[MAX_TEXT + 2];
	int iEventUnused;
	int iERoom, iETile, iENext;
	int iShowX, iShowY;
	int iXY[2 + 2];

	/*** Used for looping. ***/
	int iLoopTile;
	int iLoopRoom;
	int iLoopSide;

	/*** black background ***/
	ShowImage (imgblack, 0, 0, "imgblack", ascreen, iScale, 1);

	if (iScreenS == 1)
	{
		/*** above this room ***/
		for (iLoopTile = 1; iLoopTile <= 10; iLoopTile++)
		{
			if (arRoomLinks[iCurRoom][3] != 0)
			{
				ShowTile (arBlock[arRoomLinks[iCurRoom][3]][20 + iLoopTile],
					-11 + iLoopTile);
			} else {
// TODO show whatever is shown if there's no room above
			}
		}

		for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
		{
			LocationToXY (iLoopTile, iXY);

			ShowTile (arBlock[iCurRoom][iLoopTile], iLoopTile);
			if (iSelected == iLoopTile)
			{
				ShowImage (imgselblock, iXY[0], iXY[1], "imgselblock",
					ascreen, iScale, 1);
			}
		}

		for (iLoopTile = 1; iLoopTile <= TILES; iLoopTile++)
		{
			LocationToXY (iLoopTile, iXY);

			/*** prince ***/
			if ((iCurRoom == iKidRoom) && (iLoopTile == iKidPos))
			{
				switch (iKidDir)
				{
					case 0x00: /*** looks right ***/
						iImgX = iXY[0] + 67;
						iImgY = iXY[1] + 19;
						ShowImage (imgkidr[1], iImgX, iImgY, "imgkidr[1]",
							ascreen, iScale, 1);
						if (iSelected == iLoopTile)
						{ ShowImage (imgkidr[2], iImgX, iImgY, "imgkidr[2]",
							ascreen, iScale, 1); }
						break;
					case 0xFF: /*** looks left ***/
						iImgX = iXY[0] + 33;
						iImgY = iXY[1] + 19;
						ShowImage (imgkidl[1], iImgX, iImgY, "imgkidl[1]",
							ascreen, iScale, 1);
						if (iSelected == iLoopTile)
						{ ShowImage (imgkidl[2], iImgX, iImgY, "imgkidl[2]",
							ascreen, iScale, 1); }
						break;
					default: printf ("[FAILED] Wrong dir.!\n"); exit (EXIT_ERROR); break;
				}
			}

			/*** guard ***/
			if (arGuardLocations[iCurRoom] == iLoopTile)
			{
				if (arGuardDirections[iCurRoom] == 0xFF) /*** looks left ***/
				{
					iImgDir = 1;
					switch (iCurGuard)
					{
						case 0: /*** guard ***/
							iImgX = iXY[0] + 34; iImgY = iXY[1] + 27; break;
						case 1: /*** fat ***/
							iImgX = iXY[0] + 34; /* TODO =guess */ iImgY = iXY[1] + 26; break;
						case 2: /*** skeleton ***/
							iImgX = iXY[0] + 34; /* TODO =guess */ iImgY = iXY[1] + 31; break;
						case 3: /*** Jaffar ***/
							iImgX = iXY[0] + 34; iImgY = iXY[1] + 22; break;
						case 4: /*** shadow ***/
							iImgX = iXY[0] + 34 /* TODO =guess */; iImgY = iXY[1] + 30; break;
						default: /*** Fallback. ***/
							iImgX = iXY[0] + 34; iImgY = iXY[1] + 27; break;
					}
				} else { /*** looks right ***/
					iImgDir = 2;
					switch (iCurGuard)
					{
						case 0: /*** guard ***/
							iImgX = iXY[0] + 11; iImgY = iXY[1] + 27; break;
						case 1: /*** fat ***/
							iImgX = iXY[0] + 12; iImgY = iXY[1] + 26; break;
						case 2: /*** skeleton ***/
							iImgX = iXY[0] + 11; /* TODO =guess */ iImgY = iXY[1] + 31; break;
						case 3: /*** Jaffar ***/
							iImgX = iXY[0] + 11; /* TODO =guess */ iImgY = iXY[1] + 22; break;
						case 4: /*** shadow ***/
							iImgX = iXY[0] + 11; /* TODO =guess */ iImgY = iXY[1] + 30; break;
						default: /*** Fallback. ***/
							iImgX = iXY[0] + 11; iImgY = iXY[1] + 27; break;
					}
				}

				switch (iCurGuard)
				{
					case 0: /*** guard ***/
						// TODO, needs to be a switch, that adds imgguardpb and imgguardunk
						ShowImage (imgguardbg[iImgDir], iImgX, iImgY,
							"imgguardbg[]", ascreen, iScale, 1);
						if (iSelected == iLoopTile)
						{
							ShowImage (imgguardsel[iImgDir], iImgX, iImgY,
								"imgguardsel[]", ascreen, iScale, 1);
						}
						break;
					case 1: /*** fat ***/
						ShowImage (imgfat[iImgDir], iImgX, iImgY,
							"imgfat[]", ascreen, iScale, 1);
						if (iSelected == iLoopTile)
						{
							ShowImage (imgfatsel[iImgDir], iImgX, iImgY,
								"imgfatsel[]", ascreen, iScale, 1);
						}
						break;
					case 2: /*** skeleton ***/
						ShowImage (imgskel[iImgDir], iImgX, iImgY,
							"imgskel[]", ascreen, iScale, 1);
						if (iSelected == iLoopTile)
						{
							ShowImage (imgskelsel[iImgDir], iImgX, iImgY,
								"imgskelsel[]", ascreen, iScale, 1);
						}
						break;
					case 3: /*** Jaffar ***/
						ShowImage (imgjaffar[iImgDir], iImgX, iImgY,
							"imgjaffar[]", ascreen, iScale, 1);
						if (iSelected == iLoopTile)
						{
							ShowImage (imgjaffarsel[iImgDir], iImgX, iImgY,
								"imgjaffarsel[]", ascreen, iScale, 1);
						}
						break;
					case 4: /*** shadow ***/
						ShowImage (imgshadow[iImgDir], iImgX, iImgY,
							"imgshadow[]", ascreen, iScale, 1);
						if (iSelected == iLoopTile)
						{
							ShowImage (imgshadowsel[iImgDir], iImgX, iImgY,
								"imgshadowsel[]", ascreen, iScale, 1);
						}
						break;
				}

				switch (arGuardSkills[iCurRoom])
				{
					case 10: /*** TODO: Does PC-98 have this? ***/
						snprintf (arText[0], MAX_TEXT, "%s", "LVL:a"); break;
					case 11: /*** TODO: Does PC-98 have this? ***/
						snprintf (arText[0], MAX_TEXT, "%s", "LVL:b"); break;
					default:
						snprintf (arText[0], MAX_TEXT, "LVL:%i",
							arGuardSkills[iCurRoom]); break;
				}
				DisplayText (iXY[0], iXY[1], 11, arText, 1,
					font11, color_wh, color_bl);
			}
		}
	}
	if (iScreenS == 2) /*** R ***/
	{
		if (iBrokenRoomLinks == 0)
		{
			InitRooms();
			/*** room links ***/
			ShowImage (imgrl, 25, 50, "imgrl", ascreen, iScale, 1);
			/*** rooms broken on ***/
			if (iDownAt == 11)
			{ /*** down ***/
				ShowImage (imgbroomson_1, 622, 68, "imgbroomson_1",
					ascreen, iScale, 1);
			} else { /*** up ***/
				ShowImage (imgbroomson_0, 622, 68, "imgbroomson_0",
					ascreen, iScale, 1);
			}
			WhereToStart();
			for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
			{
				iDone[iLoopRoom] = 0;
			}
			ShowRooms (iKidRoom, iStartRoomsX, iStartRoomsY, 1);
			iUnusedRooms = 0;
			for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
			{
				if (iDone[iLoopRoom] != 1)
				{
					iUnusedRooms++;
					ShowRooms (iLoopRoom, 25, iUnusedRooms, 0);
				}
			}
			if (iMovingRoom != 0)
			{
				iXShow = (iXPos + 10) / iScale;
				iYShow = (iYPos + 10) / iScale;
				ShowImage (imgroom[iMovingRoom], iXShow, iYShow, "imgroom[]",
					ascreen, iScale, 1);
				if (iCurRoom == iMovingRoom) /*** green stripes ***/
					{ ShowImage (imgsrc, iXShow, iYShow, "imgsrc", ascreen, iScale, 1); }
				if (iKidRoom == iMovingRoom) /*** blue border ***/
					{ ShowImage (imgsrs, iXShow, iYShow, "imgsrs", ascreen, iScale, 1); }
				ShowImage (imgsrm, iXShow, iYShow, "imgsrm", ascreen, iScale, 1);
				ShowRooms (-1, iMovingNewX, iMovingNewY, 0);
			}
		} else {
			/*** broken room links ***/
			ShowImage (imgbrl, 25, 50, "imgbrl", ascreen, iScale, 1);
			for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
			{
				switch (iLoopRoom)
				{
					case 1: iXShow = (63 * 0); iYShow = (63 * 0); break;
					case 2: iXShow = (63 * 1); iYShow = (63 * 0); break;
					case 3: iXShow = (63 * 2); iYShow = (63 * 0); break;
					case 4: iXShow = (63 * 3); iYShow = (63 * 0); break;
					case 5: iXShow = (63 * 0); iYShow = (63 * 1); break;
					case 6: iXShow = (63 * 1); iYShow = (63 * 1); break;
					case 7: iXShow = (63 * 2); iYShow = (63 * 1); break;
					case 8: iXShow = (63 * 3); iYShow = (63 * 1); break;
					case 9: iXShow = (63 * 0); iYShow = (63 * 2); break;
					case 10: iXShow = (63 * 1); iYShow = (63 * 2); break;
					case 11: iXShow = (63 * 2); iYShow = (63 * 2); break;
					case 12: iXShow = (63 * 3); iYShow = (63 * 2); break;
					case 13: iXShow = (63 * 0); iYShow = (63 * 3); break;
					case 14: iXShow = (63 * 1); iYShow = (63 * 3); break;
					case 15: iXShow = (63 * 2); iYShow = (63 * 3); break;
					case 16: iXShow = (63 * 3); iYShow = (63 * 3); break;
					case 17: iXShow = (63 * 0); iYShow = (63 * 4); break;
					case 18: iXShow = (63 * 1); iYShow = (63 * 4); break;
					case 19: iXShow = (63 * 2); iYShow = (63 * 4); break;
					case 20: iXShow = (63 * 3); iYShow = (63 * 4); break;
					case 21: iXShow = (63 * 0); iYShow = (63 * 5); break;
					case 22: iXShow = (63 * 1); iYShow = (63 * 5); break;
					case 23: iXShow = (63 * 2); iYShow = (63 * 5); break;
					case 24: iXShow = (63 * 3); iYShow = (63 * 5); break;
				}
				if (iCurRoom == iLoopRoom)
				{ /*** green stripes ***/
					ShowImage (imgsrc, iXShow + broken_room_x,
						iYShow + broken_room_y, "imgsrc", ascreen, iScale, 1);
				}
				if (iKidRoom == iLoopRoom)
				{ /*** blue border ***/
					ShowImage (imgsrs, iXShow + broken_room_x,
						iYShow + broken_room_y, "imgsrs", ascreen, iScale, 1);
				}
				for (iLoopSide = 1; iLoopSide <= 4; iLoopSide++)
				{
					if (arRoomLinks[iLoopRoom][iLoopSide] != 0)
					{
						snprintf (sShowRoom, MAX_TEXT, "imgroom[%i]",
							arRoomLinks[iLoopRoom][iLoopSide]);
						ShowImage (imgroom[arRoomLinks[iLoopRoom][iLoopSide]],
							iXShow + broken_side_x[iLoopSide],
							iYShow + broken_side_y[iLoopSide],
							sShowRoom, ascreen, iScale, 1);

						if (arRoomLinksBroken[iLoopRoom][iLoopSide] == 1)
						{ /*** blue square ***/
							ShowImage (imgsrb,
								iXShow + broken_side_x[iLoopSide],
								iYShow + broken_side_y[iLoopSide],
								"imgsrb", ascreen, iScale, 1);
						}
					}

					if ((iChangingBrokenRoom == iLoopRoom) &&
						(iChangingBrokenSide == iLoopSide))
					{ /*** red stripes ***/
						ShowImage (imgsrm,
							iXShow + broken_side_x[iLoopSide],
							iYShow + broken_side_y[iLoopSide],
							"imgsrm", ascreen, iScale, 1);
					}
				}
			}
		}
	}
	if (iScreenS == 3) /*** E ***/
	{
		iEventUnused = 0;

		/*** events screen ***/
		ShowImage (imgevents, 25, 50, "imgevents", ascreen, iScale, 1);

		/*** current event number ***/
		CenterNumber (ascreen, iChangeEvent, 364, 65, color_wh, 0);

		/*** sel event, room ***/
		iERoom = EventInfo (iChangeEvent, 1);
		if ((iERoom >= 1) && (iERoom <= ROOMS))
		{
			iShowX = 275 + (15 * iERoom);
			iShowY = 120;
			ShowImage (imgsele, iShowX, iShowY, "imgsele", ascreen, iScale, 1);
		} else { iEventUnused = 1; }

		/*** sel event, tile ***/
		iETile = EventInfo (iChangeEvent, 2);
		if ((iETile >= 1) && (iETile <= TILES))
		{
			iShowX = -10; iShowY = -10; /*** To prevent warnings. ***/
			if ((iETile >= 1) && (iETile <= 10))
			{
				iShowX = 440 + (15 * iETile);
				iShowY = 160;
			}
			if ((iETile >= 11) && (iETile <= 20))
			{
				iShowX = 440 + (15 * (iETile - 10));
				iShowY = 175;
			}
			if ((iETile >= 21) && (iETile <= 30))
			{
				iShowX = 440 + (15 * (iETile - 20));
				iShowY = 190;
			}
			ShowImage (imgsele, iShowX, iShowY, "imgsele", ascreen, iScale, 1);
		} else { iEventUnused = 1; }

		/*** sel event, next ***/
		iENext = EventInfo (iChangeEvent, 3);
		if ((iENext == 0) || (iENext == 1))
		{
			switch (iENext)
			{
				case 0: iShowX = 575; break; /*** n ***/
				case 1: iShowX = 590; break; /*** y ***/
			}
			ShowImage (imgsele, iShowX, 230, "imgsele", ascreen, iScale, 1);
		}

		/*** Show target. ***/
		if (iEventUnused == 0)
		{
			ShowTile (arBlock[iERoom][iETile], 104);
		} else {
			ShowImage (imgeventu, 90, 166, "imgeventu", ascreen, iScale, 1);
		}
	}

	if (arRoomLinks[iCurRoom][1] != 0)
	{ /*** left ***/
		if (iDownAt == 1)
		{ /*** down ***/
			ShowImage (imgleft_1, 0, 50, "imgleft_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgleft_0, 0, 50, "imgleft_0", ascreen, iScale, 1);
		}
	} else { /*** no left ***/
		ShowImage (imglrno, 0, 50, "imglrno", ascreen, iScale, 1);
	}
	if (arRoomLinks[iCurRoom][2] != 0)
	{ /*** right ***/
		if (iDownAt == 2)
		{ /*** down ***/
			ShowImage (imgright_1, 665, 50, "imgright_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgright_0, 665, 50, "imgright_0", ascreen, iScale, 1);
		}
	} else { /*** no right ***/
		ShowImage (imglrno, 665, 50, "imglrno", ascreen, iScale, 1);
	}
	if (arRoomLinks[iCurRoom][3] != 0)
	{ /*** up ***/
		if (iDownAt == 3)
		{ /*** down ***/
			ShowImage (imgup_1, 25, 25, "imgup_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgup_0, 25, 25, "imgup_0", ascreen, iScale, 1);
		}
	} else { /*** no up ***/
		if (iScreen != 1)
		{ /*** without info ***/
			ShowImage (imgudno, 25, 25, "imgudno", ascreen, iScale, 1);
		} else { /*** with info ***/
			ShowImage (imgudnonfo, 25, 25, "imgudnonfo", ascreen, iScale, 1);
		}
	}
	if (arRoomLinks[iCurRoom][4] != 0)
	{ /*** down ***/
		if (iDownAt == 4)
		{ /*** down ***/
			ShowImage (imgdown_1, 25, 442, "imgdown_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgdown_0, 25, 442, "imgdown_0", ascreen, iScale, 1);
		}
	} else { /*** no down ***/
		ShowImage (imgudno, 25, 442, "imgudno", ascreen, iScale, 1);
	}

	/*** room links ***/
	if (iScreen != 2)
	{
		if (iBrokenRoomLinks == 1)
		{
			/*** broken room links ***/
			if (iDownAt == 11)
			{ /*** down ***/
				ShowImage (imgbroomson_1, 0, 25, "imgbroomson_1", ascreen, iScale, 1);
			} else { /*** up ***/
				ShowImage (imgbroomson_0, 0, 25, "imgbroomson_0", ascreen, iScale, 1);
			}
		} else {
			/*** regular room links ***/
			if (iDownAt == 5)
			{ /*** down ***/
				ShowImage (imgroomson_1, 0, 25, "imgroomson_1", ascreen, iScale, 1);
			} else { /*** up ***/
				ShowImage (imgroomson_0, 0, 25, "imgroomson_0", ascreen, iScale, 1);
			}
		}
	} else {
		if (iBrokenRoomLinks == 1)
		{
			/*** broken room links ***/
			ShowImage (imgbroomsoff, 0, 25, "imgbroomsoff", ascreen, iScale, 1);
		} else {
			/*** regular room links ***/
			ShowImage (imgroomsoff, 0, 25, "imgroomsoff", ascreen, iScale, 1);
		}
	}

	/*** events ***/
	if (iScreen != 3)
	{
		if (iDownAt == 6)
		{ /*** down ***/
			ShowImage (imgeventson_1, 665, 25, "imgeventson_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgeventson_0, 665, 25, "imgeventson_0", ascreen, iScale, 1);
		}
	} else { /*** off ***/
		ShowImage (imgeventsoff, 665, 25, "imgeventsoff", ascreen, iScale, 1);
	}

	/*** save ***/
	if (iChanged != 0)
	{ /*** on ***/
		if (iDownAt == 7)
		{ /*** down ***/
			ShowImage (imgsaveon_1, 0, 442, "imgsaveon_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgsaveon_0, 0, 442, "imgsaveon_0", ascreen, iScale, 1);
		}
	} else { /*** off ***/
		ShowImage (imgsaveoff, 0, 442, "imgsaveoff", ascreen, iScale, 1);
	}

	/*** quit ***/
	if (iDownAt == 8)
	{ /*** down ***/
		ShowImage (imgquit_1, 665, 442, "imgquit_1", ascreen, iScale, 1);
	} else { /*** up ***/
		ShowImage (imgquit_0, 665, 442, "imgquit_0", ascreen, iScale, 1);
	}

	/*** prev ***/
	if (iCurLevel != 1)
	{ /*** on ***/
		if (iDownAt == 9)
		{ /*** down ***/
			ShowImage (imgprevon_1, 0, 0, "imgprevon_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgprevon_0, 0, 0, "imgprevon_0", ascreen, iScale, 1);
		}
	} else { /*** off ***/
		ShowImage (imgprevoff, 0, 0, "imgprevoff", ascreen, iScale, 1);
	}

	/*** next ***/
	if (iCurLevel != 0)
	{ /*** on ***/
		if (iDownAt == 10)
		{ /*** down ***/
			ShowImage (imgnexton_1, 665, 0, "imgnexton_1", ascreen, iScale, 1);
		} else { /*** up ***/
			ShowImage (imgnexton_0, 665, 0, "imgnexton_0", ascreen, iScale, 1);
		}
	} else { /*** off ***/
		ShowImage (imgnextoff, 665, 0, "imgnextoff", ascreen, iScale, 1);
	}

	/*** level bar ***/
	ShowImage (imgbar, 25, 0, "imgbar", ascreen, iScale, 1);
	switch (iCurLevel)
	{
		case 0: snprintf (sLevel, MAX_TEXT, "%s", "level 0 (demo)"); break;
		case 1: snprintf (sLevel, MAX_TEXT, "%s", "level 1 (prison)"); break;
		case 2: snprintf (sLevel, MAX_TEXT, "%s", "level 2 (guards)"); break;
		case 3: snprintf (sLevel, MAX_TEXT, "%s", "level 3 (skeleton)"); break;
		case 4: snprintf (sLevel, MAX_TEXT, "%s", "level 4 (mirror)"); break;
		case 5: snprintf (sLevel, MAX_TEXT, "%s", "level 5 (thief)"); break;
		case 6: snprintf (sLevel, MAX_TEXT, "%s", "level 6 (plunge)"); break;
		case 7: snprintf (sLevel, MAX_TEXT, "%s", "level 7 (weightless)"); break;
		case 8: snprintf (sLevel, MAX_TEXT, "%s", "level 8 (mouse)"); break;
		case 9: snprintf (sLevel, MAX_TEXT, "%s", "level 9 (twisty)"); break;
		case 10: snprintf (sLevel, MAX_TEXT, "%s", "level 10 (quad)"); break;
		case 11: snprintf (sLevel, MAX_TEXT, "%s", "level 11 (fragile)"); break;
		case 12: snprintf (sLevel, MAX_TEXT, "%s", "level 12 (12a; tower)"); break;
		case 13: snprintf (sLevel, MAX_TEXT, "%s",
			"level 13 (12b; jaffar)"); break;
		case 14: snprintf (sLevel, MAX_TEXT, "%s", "level 14 (rescue)"); break;
		case 15: snprintf (sLevel, MAX_TEXT, "%s", "level 15 (potions)"); break;
	}
	switch (iScreen)
	{
		case 1:
			snprintf (arText[0], MAX_TEXT, "%s, room %i", sLevel, iCurRoom);
			ShowImage (extras[iExtras], 608, 3, "extras[iExtras]",
				ascreen, iScale, 1);
			break;
		case 2:
			snprintf (arText[0], MAX_TEXT, "%s, room links", sLevel); break;
		case 3:
			snprintf (arText[0], MAX_TEXT, "%s, events", sLevel); break;
	}
	DisplayText (31, 5, 15, arText, 1, font15, color_wh, color_bl);
	if (iPlaytest == 1)
		{ ShowImage (imgplaytest, 25, 50, "imgplaytest", ascreen, iScale, 1); }

	/*** refresh screen ***/
	SDL_RenderPresent (ascreen);
}
/*****************************************************************************/
void InitPopUp (void)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
void InitPopUpSave (void)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
void ShowTile (int iBlock, int iLocation)
/*****************************************************************************/
{
	int iXY[2 + 2];
	SDL_Texture *imgshow;
	char arText[1 + 2][MAX_TEXT + 2];
	/***/
	int iFlag, iObject;
	char sFlagName[5 + 2];
	char sObjectName[5 + 2];

	LocationToXY (iLocation, iXY);

	imgshow = imgunk;
	if ((iCurLevel >= 0) && (iCurLevel <= 3))
	{
		if (imgl0003[iBlock] != NULL)
		{
			imgshow = imgl0003[iBlock];
		}
	}
	// TODO other levels (l0406 l0709 l1011 l1213 l14 l15)
	ShowImage (imgshow, iXY[0], iXY[1], "imgshow", ascreen, iScale, 1);

	iFlag = arBlockFlag[iBlock & 0x7f];
	iObject = arBlockObject[iBlock & 0x7f];
	if ((iFlag != 0) || (iObject != 0))
	{
		GetFlagName (iFlag, sFlagName);
		GetObjectName (iObject, sObjectName);
		snprintf (arText[0], MAX_TEXT, "%s", sFlagName);
		snprintf (arText[1], MAX_TEXT, "%s", sObjectName);
		DisplayText (iXY[0] + 3, iXY[1] + 97, 11, arText,
			2, font11, color_wh, color_bl);
	}
}
/*****************************************************************************/
void LocationToXY (int iLocation, int iXY[4])
/*****************************************************************************/
{
	switch (iLocation)
	{
		/*** above this room ***/
		case -10: iXY[0] = 25+(64*0); iXY[1] = 58+(128*-1); break;
		case -9: iXY[0] = 25+(64*1); iXY[1] = 58+(128*-1); break;
		case -8: iXY[0] = 25+(64*2); iXY[1] = 58+(128*-1); break;
		case -7: iXY[0] = 25+(64*3); iXY[1] = 58+(128*-1); break;
		case -6: iXY[0] = 25+(64*4); iXY[1] = 58+(128*-1); break;
		case -5: iXY[0] = 25+(64*5); iXY[1] = 58+(128*-1); break;
		case -4: iXY[0] = 25+(64*6); iXY[1] = 58+(128*-1); break;
		case -3: iXY[0] = 25+(64*7); iXY[1] = 58+(128*-1); break;
		case -2: iXY[0] = 25+(64*8); iXY[1] = 58+(128*-1); break;
		case -1: iXY[0] = 25+(64*9); iXY[1] = 58+(128*-1); break;
		/*** this room ***/
		case 1: iXY[0] = 25+(64*0); iXY[1] = 58+(128*0); break;
		case 2: iXY[0] = 25+(64*1); iXY[1] = 58+(128*0); break;
		case 3: iXY[0] = 25+(64*2); iXY[1] = 58+(128*0); break;
		case 4: iXY[0] = 25+(64*3); iXY[1] = 58+(128*0); break;
		case 5: iXY[0] = 25+(64*4); iXY[1] = 58+(128*0); break;
		case 6: iXY[0] = 25+(64*5); iXY[1] = 58+(128*0); break;
		case 7: iXY[0] = 25+(64*6); iXY[1] = 58+(128*0); break;
		case 8: iXY[0] = 25+(64*7); iXY[1] = 58+(128*0); break;
		case 9: iXY[0] = 25+(64*8); iXY[1] = 58+(128*0); break;
		case 10: iXY[0] = 25+(64*9); iXY[1] = 58+(128*0); break;
		case 11: iXY[0] = 25+(64*0); iXY[1] = 58+(128*1); break;
		case 12: iXY[0] = 25+(64*1); iXY[1] = 58+(128*1); break;
		case 13: iXY[0] = 25+(64*2); iXY[1] = 58+(128*1); break;
		case 14: iXY[0] = 25+(64*3); iXY[1] = 58+(128*1); break;
		case 15: iXY[0] = 25+(64*4); iXY[1] = 58+(128*1); break;
		case 16: iXY[0] = 25+(64*5); iXY[1] = 58+(128*1); break;
		case 17: iXY[0] = 25+(64*6); iXY[1] = 58+(128*1); break;
		case 18: iXY[0] = 25+(64*7); iXY[1] = 58+(128*1); break;
		case 19: iXY[0] = 25+(64*8); iXY[1] = 58+(128*1); break;
		case 20: iXY[0] = 25+(64*9); iXY[1] = 58+(128*1); break;
		case 21: iXY[0] = 25+(64*0); iXY[1] = 58+(128*2); break;
		case 22: iXY[0] = 25+(64*1); iXY[1] = 58+(128*2); break;
		case 23: iXY[0] = 25+(64*2); iXY[1] = 58+(128*2); break;
		case 24: iXY[0] = 25+(64*3); iXY[1] = 58+(128*2); break;
		case 25: iXY[0] = 25+(64*4); iXY[1] = 58+(128*2); break;
		case 26: iXY[0] = 25+(64*5); iXY[1] = 58+(128*2); break;
		case 27: iXY[0] = 25+(64*6); iXY[1] = 58+(128*2); break;
		case 28: iXY[0] = 25+(64*7); iXY[1] = 58+(128*2); break;
		case 29: iXY[0] = 25+(64*8); iXY[1] = 58+(128*2); break;
		case 30: iXY[0] = 25+(64*9); iXY[1] = 58+(128*2); break;
		/*** event target ***/
		case 104: iXY[0] = 105; iXY[1] = 181; break;
		default:
			printf ("[ WARN ] Unknown tile location: %i!\n", iLocation);
			break;
	}
}
/*****************************************************************************/
int InArea (int iUpperLeftX, int iUpperLeftY,
	int iLowerRightX, int iLowerRightY)
/*****************************************************************************/
{
	if ((iUpperLeftX * iScale <= iXPos) &&
		(iLowerRightX * iScale >= iXPos) &&
		(iUpperLeftY * iScale <= iYPos) &&
		(iLowerRightY * iScale >= iYPos))
	{
		return (1);
	} else {
		return (0);
	}
}
/*****************************************************************************/
void PlaySound (char *sFile)
/*****************************************************************************/
{
	int iIndex;
	SDL_AudioSpec wave;
	Uint8 *data;
	Uint32 dlen;
	SDL_AudioCVT cvt;

	if (iNoAudio == 1) { return; }
	for (iIndex = 0; iIndex < NUM_SOUNDS; iIndex++)
	{
		if (sounds[iIndex].dpos == sounds[iIndex].dlen)
		{
			break;
		}
	}
	if (iIndex == NUM_SOUNDS) { return; }

	if (SDL_LoadWAV (sFile, &wave, &data, &dlen) == NULL)
	{
		printf ("[FAILED] Could not load %s: %s!\n", sFile, SDL_GetError());
		exit (EXIT_ERROR);
	}
	SDL_BuildAudioCVT (&cvt, wave.format, wave.channels, wave.freq, AUDIO_S16, 2,
		44100);
	/*** The "+ 1" is a workaround for SDL bug #2274. ***/
	cvt.buf = (Uint8 *)malloc (dlen * (cvt.len_mult + 1));
	memcpy (cvt.buf, data, dlen);
	cvt.len = dlen;
	SDL_ConvertAudio (&cvt);
	SDL_FreeWAV (data);

	if (sounds[iIndex].data)
	{
		free (sounds[iIndex].data);
	}
	SDL_LockAudio();
	sounds[iIndex].data = cvt.buf;
	sounds[iIndex].dlen = cvt.len_cvt;
	sounds[iIndex].dpos = 0;
	SDL_UnlockAudio();
}
/*****************************************************************************/
int BrokenRoomLinks (int iPrint)
/*****************************************************************************/
{
	int iBroken;
	int iTemp;

	for (iTemp = 1; iTemp <= ROOMS; iTemp++)
	{
		iDone[iTemp] = 0;
		arRoomLinksBroken[iTemp][1] = 0;
		arRoomLinksBroken[iTemp][2] = 0;
		arRoomLinksBroken[iTemp][3] = 0;
		arRoomLinksBroken[iTemp][4] = 0;
	}
	CheckSides (iKidRoom, 0, 0);
	iBroken = 0;

	for (iTemp = 1; iTemp <= ROOMS; iTemp++)
	{
		/*** If the room is in use... ***/
		if (iDone[iTemp] == 1)
		{
			/*** check left ***/
			if (arRoomLinks[iTemp][1] != 0)
			{
				if ((arRoomLinks[iTemp][1] == iTemp) ||
					(arRoomLinks[arRoomLinks[iTemp][1]][2] != iTemp))
				{
					arRoomLinksBroken[iTemp][1] = 1;
					iBroken = 1;
					if ((iDebug == 1) && (iPrint == 1))
					{
						printf ("[ INFO ] The left of room %i has a broken link.\n",
							iTemp);
					}
				}
			}
			/*** check right ***/
			if (arRoomLinks[iTemp][2] != 0)
			{
				if ((arRoomLinks[iTemp][2] == iTemp) ||
					(arRoomLinks[arRoomLinks[iTemp][2]][1] != iTemp))
				{
					arRoomLinksBroken[iTemp][2] = 1;
					iBroken = 1;
					if ((iDebug == 1) && (iPrint == 1))
					{
						printf ("[ INFO ] The right of room %i has a broken link.\n",
							iTemp);
					}
				}
			}
			/*** check up ***/
			if (arRoomLinks[iTemp][3] != 0)
			{
				if ((arRoomLinks[iTemp][3] == iTemp) ||
					(arRoomLinks[arRoomLinks[iTemp][3]][4] != iTemp))
				{
					arRoomLinksBroken[iTemp][3] = 1;
					iBroken = 1;
					if ((iDebug == 1) && (iPrint == 1))
					{
						printf ("[ INFO ] The top of room %i has a broken link.\n",
							iTemp);
					}
				}
			}
			/*** check down ***/
			if (arRoomLinks[iTemp][4] != 0)
			{
				if ((arRoomLinks[iTemp][4] == iTemp) ||
					(arRoomLinks[arRoomLinks[iTemp][4]][3] != iTemp))
				{
					arRoomLinksBroken[iTemp][4] = 1;
					iBroken = 1;
					if ((iDebug == 1) && (iPrint == 1))
					{
						printf ("[ INFO ] The bottom of room %i has a broken link.\n",
							iTemp);
					}
				}
			}
		}
	}

	return (iBroken);
}
/*****************************************************************************/
void CheckSides (int iRoom, int iX, int iY)
/*****************************************************************************/
{
	if (iX < iMinX) { iMinX = iX; }
	if (iY < iMinY) { iMinY = iY; }
	if (iX > iMaxX) { iMaxX = iX; }
	if (iY > iMaxY) { iMaxY = iY; }

	iDone[iRoom] = 1;

	if ((arRoomLinks[iRoom][1] != 0) &&
		(iDone[arRoomLinks[iRoom][1]] != 1)) 
		{ CheckSides (arRoomLinks[iRoom][1], iX - 1, iY); }

	if ((arRoomLinks[iRoom][2] != 0) &&
		(iDone[arRoomLinks[iRoom][2]] != 1)) 
		{ CheckSides (arRoomLinks[iRoom][2], iX + 1, iY); }

	if ((arRoomLinks[iRoom][3] != 0) &&
		(iDone[arRoomLinks[iRoom][3]] != 1)) 
		{ CheckSides (arRoomLinks[iRoom][3], iX, iY - 1); }

	if ((arRoomLinks[iRoom][4] != 0) &&
		(iDone[arRoomLinks[iRoom][4]] != 1)) 
		{ CheckSides (arRoomLinks[iRoom][4], iX, iY + 1); }
}
/*****************************************************************************/
void Prev (void)
/*****************************************************************************/
{
	int iToLoad;

	if (iCurLevel != 1)
	{
		iToLoad = iCurLevel - 1;
		if (iToLoad == -1) { iToLoad = MAX_LEVEL; }
		LoadLevel (iToLoad);
		ShowScreen (iScreen);
		PlaySound ("wav/level_change.wav");
	}
}
/*****************************************************************************/
void Next (void)
/*****************************************************************************/
{
	int iToLoad;

	if (iCurLevel != 0)
	{
		iToLoad = iCurLevel + 1;
		if (iToLoad == 16) { iToLoad = 0; }
		LoadLevel (iToLoad);
		ShowScreen (iScreen);
		PlaySound ("wav/level_change.wav");
	}
}
/*****************************************************************************/
void DisplayText (int iStartX, int iStartY, int iFontSize,
	char arText[9 + 2][MAX_TEXT + 2], int iLines, TTF_Font *font,
	SDL_Color back, SDL_Color fore)
/*****************************************************************************/
{
	/*** Used for looping. ***/
	int iLoopLine;

	for (iLoopLine = 0; iLoopLine < iLines; iLoopLine++)
	{
		if (strcmp (arText[iLoopLine], "") != 0)
		{
			message = TTF_RenderText_Shaded (font,
				arText[iLoopLine], fore, back);
			messaget = SDL_CreateTextureFromSurface (ascreen, message);
			if ((strcmp (arText[iLoopLine], "single tile (change or select)") == 0)
				|| (strcmp (arText[iLoopLine], "entire room (clear or fill)") == 0) ||
				(strcmp (arText[iLoopLine], "entire level (randomize or fill)") == 0))
			{
				offset.x = iStartX + 20;
			} else {
				offset.x = iStartX;
			}
			offset.y = iStartY + (iLoopLine * (iFontSize + 4));
			offset.w = message->w; offset.h = message->h;
			CustomRenderCopy (messaget, "message", NULL, ascreen, &offset);
			SDL_DestroyTexture (messaget); SDL_FreeSurface (message);
		}
	}
}
/*****************************************************************************/
void CustomRenderCopy (SDL_Texture* src, char *sImageInfo, SDL_Rect* srcrect,
	SDL_Renderer* dst, SDL_Rect *dstrect)
/*****************************************************************************/
{
	SDL_Rect stuff;

	stuff.x = dstrect->x * iScale;
	stuff.y = dstrect->y * iScale;
	if (srcrect != NULL) /*** image ***/
	{
		stuff.w = dstrect->w * iScale;
		stuff.h = dstrect->h * iScale;
	} else { /*** font ***/
		stuff.w = dstrect->w;
		stuff.h = dstrect->h;
	}
	if (SDL_RenderCopy (dst, src, srcrect, &stuff) != 0)
	{
		printf ("[ WARN ] SDL_RenderCopy (%s): %s!\n",
			sImageInfo, SDL_GetError());
	}
}
/*****************************************************************************/
void ChangeEvent (int iAmount, int iChangePos)
/*****************************************************************************/
{
	if (((iAmount > 0) && (iChangeEvent != EVENTS)) ||
		((iAmount < 0) && (iChangeEvent != 1)))
	{
		switch (iAmount)
		{
			case -10: iChangeEvent-=10; break;
			case -1: iChangeEvent--; break;
			case 1: iChangeEvent++; break;
			case 10: iChangeEvent+=10; break;
		}
		if (iChangeEvent < 1) { iChangeEvent = 1; }
		if (iChangeEvent > EVENTS) { iChangeEvent = EVENTS; }
		if (iChangePos == 1)
		{
			if ((arBlockObject[arBlock[iCurRoom][iSelected]] == 0x10) ||
				(arBlockObject[arBlock[iCurRoom][iSelected]] == 0x18))
			{
				arModif[iCurRoom][iSelected] = iChangeEvent;
				iChanged++;
			}
		}
		PlaySound ("wav/plus_minus.wav");
	}
}
/*****************************************************************************/
int OnLevelBar (void)
/*****************************************************************************/
{
	if (InArea (28, 3, 28 + 575, 3 + 19) == 1)
		{ return (1); } else { return (0); }
}
/*****************************************************************************/
void InitRooms (void)
/*****************************************************************************/
{
	/*** Used for looping. ***/
	int iLoopX;
	int iLoopY;

	for (iLoopX = 0; iLoopX <= ROOMS + 1; iLoopX++)
	{
		for (iLoopY = 0; iLoopY <= ROOMS; iLoopY++)
		{
			arRoomMoving[iLoopX][iLoopY] = 0;
		}
	}
}
/*****************************************************************************/
void WhereToStart (void)
/*****************************************************************************/
{
	/*** Used for looping. ***/
	int iLoopRoom;

	iMinX = 0;
	iMaxX = 0;
	iMinY = 0;
	iMaxY = 0;

	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		iDone[iLoopRoom] = 0;
	}
	CheckSides (iKidRoom, 0, 0);

	iStartRoomsX = round (12 - (((float)iMinX + (float)iMaxX) / 2));
	iStartRoomsY = round (12 - (((float)iMinY + (float)iMaxY) / 2));
}
/*****************************************************************************/
void ShowRooms (int iRoom, int iX, int iY, int iNext)
/*****************************************************************************/
{
	int iXShow, iYShow;
	char sShowRoom[MAX_TEXT + 2];

	if (iX == 25)
	{
		iXShow = 267;
	} else {
		iXShow = 275 + (iX * 15);
	}
	iYShow = 51 + (iY * 15);

	if (iRoom != -1)
	{
		snprintf (sShowRoom, MAX_TEXT, "imgroom[%i]", iRoom);
		ShowImage (imgroom[iRoom], iXShow, iYShow, sShowRoom,
			ascreen, iScale, 1);
		arRoomMoving[iX][iY] = iRoom;
		if (iCurRoom == iRoom) /*** green stripes ***/
			{ ShowImage (imgsrc, iXShow, iYShow, "imgsrc", ascreen, iScale, 1); }
		if (iKidRoom == iRoom) /*** blue border ***/
			{ ShowImage (imgsrs, iXShow, iYShow, "imgsrs", ascreen, iScale, 1); }
		if (iRoom == iMovingRoom) /*** red stripes ***/
			{ ShowImage (imgsrm, iXShow, iYShow, "imgsrm", ascreen, iScale, 1); }
	} else {
		/*** white cross ***/
		ShowImage (imgsrp, iXShow, iYShow, "imgsrp", ascreen, iScale, 1);
	}
	if (iRoom == iMovingRoom)
	{
		iMovingOldX = iX;
		iMovingOldY = iY;
		if (iMovingNewBusy == 0)
		{
			iMovingNewX = iMovingOldX;
			iMovingNewY = iMovingOldY;
			iMovingNewBusy = 1;
		}
	}

	if (iRoom != -1) { iDone[iRoom] = 1; }

	if (iNext == 1)
	{
		if ((arRoomLinks[iRoom][1] != 0) &&
			(iDone[arRoomLinks[iRoom][1]] != 1))
			{ ShowRooms (arRoomLinks[iRoom][1], iX - 1, iY, 1); }

		if ((arRoomLinks[iRoom][2] != 0) &&
			(iDone[arRoomLinks[iRoom][2]] != 1))
			{ ShowRooms (arRoomLinks[iRoom][2], iX + 1, iY, 1); }

		if ((arRoomLinks[iRoom][3] != 0) &&
			(iDone[arRoomLinks[iRoom][3]] != 1))
			{ ShowRooms (arRoomLinks[iRoom][3], iX, iY - 1, 1); }

		if ((arRoomLinks[iRoom][4] != 0) &&
			(iDone[arRoomLinks[iRoom][4]] != 1))
			{ ShowRooms (arRoomLinks[iRoom][4], iX, iY + 1, 1); }
	}
}
/*****************************************************************************/
void CenterNumber (SDL_Renderer *screen, int iNumber, int iX, int iY,
	SDL_Color fore, int iHex)
/*****************************************************************************/
{
	char sText[MAX_TEXT + 2];

	if (iHex == 0)
	{
		if (iNumber <= 999)
		{
			snprintf (sText, MAX_TEXT, "%i", iNumber);
		} else {
			snprintf (sText, MAX_TEXT, "%s", "X");
		}
	} else {
		snprintf (sText, MAX_TEXT, "%02X", iNumber);
	}
	message = TTF_RenderText_Blended_Wrapped (font20, sText, fore, 1000); /*** TODO: 1000 ***/
	messaget = SDL_CreateTextureFromSurface (screen, message);
	if (iHex == 0)
	{
		if ((iNumber >= -9) && (iNumber <= -1))
		{
			offset.x = iX + 16;
		} else if ((iNumber >= 0) && (iNumber <= 9)) {
			offset.x = iX + 21;
		} else if ((iNumber >= 10) && (iNumber <= 99)) {
			offset.x = iX + 14;
		} else if ((iNumber >= 100) && (iNumber <= 999)) {
			offset.x = iX + 7;
		} else {
			offset.x = iX + 21;
		}
	} else {
		offset.x = iX + 14;
	}
	offset.y = iY - 1;
	offset.w = message->w; offset.h = message->h;
	CustomRenderCopy (messaget, "message", NULL, screen, &offset);
	SDL_DestroyTexture (messaget); SDL_FreeSurface (message);
}
/*****************************************************************************/
int EventInfo (int iNr, int iType)
/*****************************************************************************/
{
	switch (iType)
	{
		case 1: /*** room ***/
			return (arEventRoom[iNr]); break;
		case 2: /*** tile ***/
			return (arEventTile[iNr]); break;
		case 3: /*** next ***/
			return (arEventNext[iNr]); break;
		default:
			printf ("[ WARN ] Strange event info type: %i!\n", iType);
			break;
	}

	return (-1);
}
/*****************************************************************************/
int BitFromByte (int iByte, int iGetBitNr)
/*****************************************************************************/
{
	/*** Returns 0 or 1. ***/

	int iBit;

	iBit = (iByte >> iGetBitNr) & 1;

	return (iBit);
}
/*****************************************************************************/
int BitsToInt (char *sString)
/*****************************************************************************/
{
	/*** Converts binary to decimal. ***/
	/*** Example: 11111111 to 255 ***/

	int iTemp = 0;

	for (; *sString; iTemp = (iTemp << 1) | (*sString++ - '0'));
	return (iTemp);
}
/*****************************************************************************/
void EventRoom (int iRoom)
/*****************************************************************************/
{
	arEventRoom[iChangeEvent] = iRoom;
	PlaySound ("wav/check_box.wav");
	iChanged++;
}
/*****************************************************************************/
void EventDoor (int iTile)
/*****************************************************************************/
{
	arEventTile[iChangeEvent] = iTile;
	PlaySound ("wav/check_box.wav");
	iChanged++;
}
/*****************************************************************************/
void EventNext (int iNoNext)
/*****************************************************************************/
{
	arEventNext[iChangeEvent] = iNoNext;
	PlaySound ("wav/check_box.wav");
	iChanged++;
}
/*****************************************************************************/
void ClearRoom (void)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
void UseTile (int iTile, int iLocation, int iRoom)
/*****************************************************************************/
{
	if ((iTile == 0) || (iLocation == 0) || (iRoom == 0)) { } // TODO

	/*** TODO ***/
	switch (iTile)
	{
		case -2: /*** custom ***/
			SetLocation (iRoom, iLocation, iCustomBlock, arModif[iRoom][iLocation]);
			break;
		case 128: /*** prince right ***/
			if ((iKidRoom != iCurRoom) || (iKidPos != iLocation) ||
				(iKidDir != 0x00))
			{
				iKidRoom = iCurRoom;
				iKidPos = iLocation;
				iKidDir = 0x00;
				PlaySound ("wav/hum_adj.wav");
			}
			break;
		case 129: /*** prince left ***/
			if ((iKidRoom != iCurRoom) || (iKidPos != iLocation) ||
				(iKidDir != 0xFF))
			{
				iKidRoom = iCurRoom;
				iKidPos = iLocation;
				iKidDir = 0xFF;
				PlaySound ("wav/hum_adj.wav");
			}
			break;
	}
}
/*****************************************************************************/
int MouseSelectAdj (void)
/*****************************************************************************/
{
	int iOnAdj;
	int iAdjBaseX;
	int iAdjBaseY;

	/*** Used for looping. ***/
	int iLoopRoom;

	iOnAdj = 0;

	for (iLoopRoom = 1; iLoopRoom <= ROOMS; iLoopRoom++)
	{
		switch (iLoopRoom)
		{
			case 1: iAdjBaseX = ADJ_BASE_X + (63 * 0);
				iAdjBaseY = ADJ_BASE_Y + (63 * 0); break;
			case 2: iAdjBaseX = ADJ_BASE_X + (63 * 1);
				iAdjBaseY = ADJ_BASE_Y + (63 * 0); break;
			case 3: iAdjBaseX = ADJ_BASE_X + (63 * 2);
				iAdjBaseY = ADJ_BASE_Y + (63 * 0); break;
			case 4: iAdjBaseX = ADJ_BASE_X + (63 * 3);
				iAdjBaseY = ADJ_BASE_Y + (63 * 0); break;
			case 5: iAdjBaseX = ADJ_BASE_X + (63 * 0);
				iAdjBaseY = ADJ_BASE_Y + (63 * 1); break;
			case 6: iAdjBaseX = ADJ_BASE_X + (63 * 1);
				iAdjBaseY = ADJ_BASE_Y + (63 * 1); break;
			case 7: iAdjBaseX = ADJ_BASE_X + (63 * 2);
				iAdjBaseY = ADJ_BASE_Y + (63 * 1); break;
			case 8: iAdjBaseX = ADJ_BASE_X + (63 * 3);
				iAdjBaseY = ADJ_BASE_Y + (63 * 1); break;
			case 9: iAdjBaseX = ADJ_BASE_X + (63 * 0);
				iAdjBaseY = ADJ_BASE_Y + (63 * 2); break;
			case 10: iAdjBaseX = ADJ_BASE_X + (63 * 1);
				iAdjBaseY = ADJ_BASE_Y + (63 * 2); break;
			case 11: iAdjBaseX = ADJ_BASE_X + (63 * 2);
				iAdjBaseY = ADJ_BASE_Y + (63 * 2); break;
			case 12: iAdjBaseX = ADJ_BASE_X + (63 * 3);
				iAdjBaseY = ADJ_BASE_Y + (63 * 2); break;
			case 13: iAdjBaseX = ADJ_BASE_X + (63 * 0);
				iAdjBaseY = ADJ_BASE_Y + (63 * 3); break;
			case 14: iAdjBaseX = ADJ_BASE_X + (63 * 1);
				iAdjBaseY = ADJ_BASE_Y + (63 * 3); break;
			case 15: iAdjBaseX = ADJ_BASE_X + (63 * 2);
				iAdjBaseY = ADJ_BASE_Y + (63 * 3); break;
			case 16: iAdjBaseX = ADJ_BASE_X + (63 * 3);
				iAdjBaseY = ADJ_BASE_Y + (63 * 3); break;
			case 17: iAdjBaseX = ADJ_BASE_X + (63 * 0);
				iAdjBaseY = ADJ_BASE_Y + (63 * 4); break;
			case 18: iAdjBaseX = ADJ_BASE_X + (63 * 1);
				iAdjBaseY = ADJ_BASE_Y + (63 * 4); break;
			case 19: iAdjBaseX = ADJ_BASE_X + (63 * 2);
				iAdjBaseY = ADJ_BASE_Y + (63 * 4); break;
			case 20: iAdjBaseX = ADJ_BASE_X + (63 * 3);
				iAdjBaseY = ADJ_BASE_Y + (63 * 4); break;
			case 21: iAdjBaseX = ADJ_BASE_X + (63 * 0);
				iAdjBaseY = ADJ_BASE_Y + (63 * 5); break;
			case 22: iAdjBaseX = ADJ_BASE_X + (63 * 1);
				iAdjBaseY = ADJ_BASE_Y + (63 * 5); break;
			case 23: iAdjBaseX = ADJ_BASE_X + (63 * 2);
				iAdjBaseY = ADJ_BASE_Y + (63 * 5); break;
			case 24: iAdjBaseX = ADJ_BASE_X + (63 * 3);
				iAdjBaseY = ADJ_BASE_Y + (63 * 5); break;
			default:
				printf ("[FAILED] iLoopRoom is not in the 1-24 range!\n");
				exit (EXIT_ERROR);
		}
		if (InArea (iAdjBaseX + 1, iAdjBaseY + 16,
			iAdjBaseX + 15, iAdjBaseY + 30) == 1)
		{
			iChangingBrokenRoom = iLoopRoom;
			iChangingBrokenSide = 1; /*** left ***/
			iOnAdj = 1;
		}
		if (InArea (iAdjBaseX + 31, iAdjBaseY + 16,
			iAdjBaseX + 45, iAdjBaseY + 30) == 1)
		{
			iChangingBrokenRoom = iLoopRoom;
			iChangingBrokenSide = 2; /*** right ***/
			iOnAdj = 1;
		}
		if (InArea (iAdjBaseX + 16, iAdjBaseY + 1,
			iAdjBaseX + 30, iAdjBaseY + 14) == 1)
		{
			iChangingBrokenRoom = iLoopRoom;
			iChangingBrokenSide = 3; /*** up ***/
			iOnAdj = 1;
		}
		if (InArea (iAdjBaseX + 16, iAdjBaseY + 31,
			iAdjBaseX + 30, iAdjBaseY + 45) == 1)
		{
			iChangingBrokenRoom = iLoopRoom;
			iChangingBrokenSide = 4; /*** down ***/
			iOnAdj = 1;
		}
	}

	return (iOnAdj);
}
/*****************************************************************************/
void ChangePos (int iLocation)
/*****************************************************************************/
{
	int iChanging;
	SDL_Event event;
	int iXPosOld, iYPosOld;
	int iUseTile;
	int iNowOn;
	int iChangeGuard;

	iChanging = 1;
	iGuardSkill = 0;
	/***/
	iCustomBlock = arBlock[iCurRoom][iLocation];
	iCustomHover = 0;

	ShowChange (iLocation);
	while (iChanging == 1)
	{
		while (SDL_PollEvent (&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					iChangeGuard = 0;
					switch (event.key.keysym.sym)
					{
						case SDLK_ESCAPE:
						case SDLK_q:
						case SDLK_c:
							iChanging = 0; break;
					}
					if (iChangeGuard == 1)
					{
						ApplySkillIfNecessary (iSelected);
						PlaySound ("wav/check_box.wav");
					}
					ShowChange (iLocation);
					break;
				case SDL_MOUSEMOTION:
					iXPosOld = iXPos;
					iYPosOld = iYPos;
					iXPos = event.motion.x;
					iYPos = event.motion.y;
					if ((iXPosOld == iXPos) && (iYPosOld == iYPos)) { break; }

					/*** custom hover ***/
					iCustomHoverOld = iCustomHover;
					if (InArea (516, 402, 516 + 132, 402 + 29) == 1)
						{ iCustomHover = 1; } else { iCustomHover = 0; }
					if (iCustomHover != iCustomHoverOld) { ShowChange (iLocation); }

					iNowOn = OnTile();
					if ((iOnTile != iNowOn) && (iNowOn != 0))
					{
						if (IsDisabled (iNowOn) == 0) { iOnTile = iNowOn; }
					}

					ShowChange (iLocation);
					break;
				case SDL_MOUSEBUTTONDOWN:
					if (event.button.button == 1)
					{
						if (InArea (654, 0, 654 + 36, 0 + 467) == 1) /*** close ***/
							{ iCloseOn = 1; }
					}
					ShowChange (iLocation);
					break;
				case SDL_MOUSEBUTTONUP:
					iCloseOn = 0;

					/*** On tile or living. ***/
					iUseTile = 0;
					if (InArea (447, 401, 447 + 66, 401 + 61) == 1) { iUseTile = 1; }
					// ^ TODO

					/*** On the custom tile area. ***/
					if (InArea (516, 402, 516 + 132, 402 + 29) == 1)
					{
						iOnTile = -2;
						iUseTile = 1;
					}

					if (event.button.button == 1) /*** left mouse button ***/
					{
						/*** Changing the custom tile. ***/
						if (InArea (526, 436, 526 + 13, 436 + 20) == 1)
							{ ChangeCustom (-16); }
						if (InArea (541, 436, 541 + 13, 436 + 20) == 1)
							{ ChangeCustom (-1); }
						if (InArea (611, 436, 611 + 13, 436 + 20) == 1)
							{ ChangeCustom (1); }
						if (InArea (626, 436, 626 + 13, 436 + 20) == 1)
							{ ChangeCustom (16); }

						if (InArea (654, 0, 654 + 36, 0 + 467) == 1) /*** close ***/
							{ iChanging = 0; }

						if (iUseTile == 1)
						{
							UseTile (iOnTile, iLocation, iCurRoom);
							if (iOnTile == -2) { iChanging = 0; } // TODO
							iChanged++;
						}
					}

					ShowChange (iLocation);
					break;
				case SDL_WINDOWEVENT:
					switch (event.window.event)
					{
						case SDL_WINDOWEVENT_EXPOSED: ShowChange (iLocation); break;
						case SDL_WINDOWEVENT_CLOSE: Quit(); break;
						case SDL_WINDOWEVENT_FOCUS_GAINED:
							iActiveWindowID = iWindowID; break;
					}
					break;
				case SDL_QUIT: Quit(); break;
			}
		}

		/*** prevent CPU eating ***/
		gamespeed = REFRESH;
		while ((SDL_GetTicks() - looptime) < gamespeed)
		{
			SDL_Delay (10);
		}
		looptime = SDL_GetTicks();
	}
	PlaySound ("wav/ok_close.wav");
}
/*****************************************************************************/
void ShowChange (int iLocation)
/*****************************************************************************/
{
	if ((iCurLevel >= 0) && (iCurLevel <= 3))
	{
		ShowImage (imgl0003t, 0, 0, "imgl0003t", ascreen, iScale, 1);
	} else if ((iCurLevel >= 4) && (iCurLevel <= 6)) {
		ShowImage (imgl0406t, 0, 0, "imgl0406t", ascreen, iScale, 1);
	} else if ((iCurLevel >= 7) && (iCurLevel <= 9)) {
		ShowImage (imgl0709t, 0, 0, "imgl0709t", ascreen, iScale, 1);
	} else if ((iCurLevel == 10) || (iCurLevel == 11)) {
		ShowImage (imgl1011t, 0, 0, "imgl1011t", ascreen, iScale, 1);
	} else if ((iCurLevel == 12) || (iCurLevel == 13)) {
		ShowImage (imgl1213t, 0, 0, "imgl1213t", ascreen, iScale, 1);
	} else if (iCurLevel == 14) {
		ShowImage (imgl14t, 0, 0, "imgl14t", ascreen, iScale, 1);
	} else if (iCurLevel == 15) {
		ShowImage (imgl15t, 0, 0, "imgl15t", ascreen, iScale, 1);
	}

	/*** close ***/
	if (iCloseOn == 0)
	{ /*** off ***/
		ShowImage (imgclosebig_0, 654, 0, "imgclosebig_0", ascreen, iScale, 1);
	} else { /*** on ***/
		ShowImage (imgclosebig_1, 654, 0, "imgclosebig_1", ascreen, iScale, 1);
	}

	/*** kid ***/
	if ((iCurRoom == iKidRoom) && (iLocation == iKidPos))
	{
		if (iKidDir == 0x00)
		{ /*** right ***/
			ShowImage (imgbordersl, 445, 399, "imgbordersl", ascreen, iScale, 1);
		} else { /*** left ***/
			ShowImage (imgbordersl, 479, 399, "imgbordersl", ascreen, iScale, 1);
		}
	}

	/*** custom tile ***/
	CenterNumber (ascreen, iCustomBlock, 554, 436, color_bl, 1);

	if (iCustomHover == 1)
		{ ShowImage (imgchover, 516, 402, "imgchover", ascreen, iScale, 1); }

	/*** refresh screen ***/
	SDL_RenderPresent (ascreen);
}
/*****************************************************************************/
void RemoveOldRoom (void)
/*****************************************************************************/
{
	arRoomMoving[iMovingOldX][iMovingOldY] = 0;

	/* Change the links of the rooms around
	 * the removed room.
	 */

	/*** left of removed ***/
	if ((iMovingOldX >= 2) && (iMovingOldX <= 24))
	{
		if (arRoomMoving[iMovingOldX - 1][iMovingOldY] != 0)
		{
			arRoomLinks[arRoomMoving[iMovingOldX - 1]
				[iMovingOldY]][2] = 0; /*** remove right ***/
		}
	}

	/*** right of removed ***/
	if ((iMovingOldX >= 1) && (iMovingOldX <= 23))
	{
		if (arRoomMoving[iMovingOldX + 1][iMovingOldY] != 0)
		{
			arRoomLinks[arRoomMoving[iMovingOldX + 1]
				[iMovingOldY]][1] = 0; /*** remove left ***/
		}
	}

	/*** above removed ***/
	if ((iMovingOldY >= 2) && (iMovingOldY <= 24))
	{
		if (arRoomMoving[iMovingOldX][iMovingOldY - 1] != 0)
		{
			arRoomLinks[arRoomMoving[iMovingOldX]
				[iMovingOldY - 1]][4] = 0; /*** remove below ***/
		}
	}

	/*** below removed ***/
	if ((iMovingOldY >= 1) && (iMovingOldY <= 23))
	{
		if (arRoomMoving[iMovingOldX][iMovingOldY + 1] != 0)
		{
			arRoomLinks[arRoomMoving[iMovingOldX]
				[iMovingOldY + 1]][3] = 0; /*** remove above ***/
		}
	}
}
/*****************************************************************************/
void AddNewRoom (int iX, int iY, int iRoom)
/*****************************************************************************/
{
	arRoomMoving[iX][iY] = iRoom;

	/* Change the links of the rooms around
	 * the new room and the room itself.
	 */

	arRoomLinks[iRoom][1] = 0;
	arRoomLinks[iRoom][2] = 0;
	arRoomLinks[iRoom][3] = 0;
	arRoomLinks[iRoom][4] = 0;
	if ((iX >= 2) && (iX <= 24)) /*** left of added ***/
	{
		if (arRoomMoving[iX - 1][iY] != 0)
		{
			arRoomLinks[arRoomMoving[iX - 1]
				[iY]][2] = iRoom; /*** add room right ***/
			arRoomLinks[iRoom][1] = arRoomMoving[iX - 1][iY];
		}
	}
	if ((iX >= 1) && (iX <= 23)) /*** right of added ***/
	{
		if (arRoomMoving[iX + 1][iY] != 0)
		{
			arRoomLinks[arRoomMoving[iX + 1]
				[iY]][1] = iRoom; /*** add room left ***/
			arRoomLinks[iRoom][2] = arRoomMoving[iX + 1][iY];
		}
	}
	if ((iY >= 2) && (iY <= 24)) /*** above added ***/
	{
		if (arRoomMoving[iX][iY - 1] != 0)
		{
			arRoomLinks[arRoomMoving[iX]
				[iY - 1]][4] = iRoom; /*** add room below ***/
			arRoomLinks[iRoom][3] = arRoomMoving[iX][iY - 1];
		}
	}
	if ((iY >= 1) && (iY <= 23)) /*** below added ***/
	{
		if (arRoomMoving[iX][iY + 1] != 0)
		{
			arRoomLinks[arRoomMoving[iX]
				[iY + 1]][3] = iRoom; /*** add room above ***/
			arRoomLinks[iRoom][4] = arRoomMoving[iX][iY + 1];
		}
	}
	PlaySound ("wav/move_room.wav");
}
/*****************************************************************************/
void LinkPlus (void)
/*****************************************************************************/
{
	int iCurrent, iNew;

	iCurrent = arRoomLinks[iChangingBrokenRoom][iChangingBrokenSide];
	if (iCurrent == ROOMS) { iNew = 0; } else { iNew = iCurrent + 1; }
	arRoomLinks[iChangingBrokenRoom][iChangingBrokenSide] = iNew;
	iChanged++;
	iBrokenRoomLinks = BrokenRoomLinks (0);
	PlaySound ("wav/hum_adj.wav");
}
/*****************************************************************************/
void LinkMinus (void)
/*****************************************************************************/
{
	int iCurrent, iNew;

	iCurrent = arRoomLinks[iChangingBrokenRoom][iChangingBrokenSide];
	if (iCurrent == 0) { iNew = ROOMS; } else { iNew = iCurrent - 1; }
	arRoomLinks[iChangingBrokenRoom][iChangingBrokenSide] = iNew;
	iChanged++;
	iBrokenRoomLinks = BrokenRoomLinks (0);
	PlaySound ("wav/hum_adj.wav");
}
/*****************************************************************************/
void CopyPaste (int iRoom, int iAction)
/*****************************************************************************/
{
	if ((iRoom == 0) || (iAction == 0)) { } // TODO
}
/*****************************************************************************/
void FlipRoom (int iRoom, int iAxis)
/*****************************************************************************/
{
	if ((iRoom == 0) || (iAxis == 0)) { } // TODO
}
/*****************************************************************************/
void RunLevel (int iLevel)
/*****************************************************************************/
{
	SDL_Thread *princethread;

	ModifyForPlaytest (iLevel);
	princethread = SDL_CreateThread (StartGame, "StartGame", NULL);
	if (princethread == NULL)
		{ printf ("[ WARN ] Could not create thread!\n"); }
}
/*****************************************************************************/
int StartGame (void *unused)
/*****************************************************************************/
{
	if (unused != NULL) { } /*** To prevent warnings. ***/

	PlaySound ("wav/playtest.wav");
	if (system (HERE BATCH_FILE) == -1)
		{ printf ("[ WARN ] Could not execute batch file!\n"); }
	if (iModified == 1) { ModifyBack(); }

	return (EXIT_NORMAL);
}
/*****************************************************************************/
void ModifyForPlaytest (int iLevel)
/*****************************************************************************/
{
	if (iLevel == 0) { } // TODO
}
/*****************************************************************************/
void ModifyBack (void)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
void Zoom (int iToggleFull)
/*****************************************************************************/
{
	if (iToggleFull == 1)
	{
		if (iFullscreen == 0)
			{ iFullscreen = SDL_WINDOW_FULLSCREEN_DESKTOP; }
				else { iFullscreen = 0; }
	} else {
		if (iFullscreen == SDL_WINDOW_FULLSCREEN_DESKTOP)
		{
			iFullscreen = 0;
			iScale = 1;
		} else if (iScale == 1) {
			iScale = 2;
		} else if (iScale == 2) {
			iFullscreen = SDL_WINDOW_FULLSCREEN_DESKTOP;
		} else {
			printf ("[ WARN ] Unknown window state!\n");
		}
	}

	SDL_SetWindowFullscreen (window, iFullscreen);
	SDL_SetWindowSize (window, (WINDOW_WIDTH) * iScale,
		(WINDOW_HEIGHT) * iScale);
	SDL_RenderSetLogicalSize (ascreen, (WINDOW_WIDTH) * iScale,
		(WINDOW_HEIGHT) * iScale);
	SDL_SetWindowPosition (window, SDL_WINDOWPOS_CENTERED,
		SDL_WINDOWPOS_CENTERED);
	TTF_CloseFont (font11);
	TTF_CloseFont (font15);
	TTF_CloseFont (font20);
	LoadFonts();
}
/*****************************************************************************/
void Help (void)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
void Sprinkle (void)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
void EXE (void)
/*****************************************************************************/
{
	// TODO
}
/*****************************************************************************/
void SetLocation (int iRoom, int iTile, int iBlock, int iModif)
/*****************************************************************************/
{
	arBlock[iRoom][iTile] = iBlock;
	arModif[iRoom][iTile] = iModif;

	iLastBlock = iBlock;
	iLastModif = iModif;
}
/*****************************************************************************/
void CreateDir (char *sDir)
/*****************************************************************************/
{
	struct stat stStatus;

	if (stat (sDir, &stStatus) == -1)
	{
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
mkdir (sDir);
#else
mkdir (sDir, 0700);
#endif
	}
}
/*****************************************************************************/
void ExtractAllFiles (void)
/*****************************************************************************/
{
	char sFileName[MAX_FILENAME + 2];

	/*** Used for looping. ***/
	int iLoopEntry;

	CreateDir (DIR_FILES);
	LoadFileNamesAndFirstSectors (iFdA);
	for (iLoopEntry = 1; iLoopEntry <= ENTRIES; iLoopEntry++)
	{
		if (iFirstSectors[iLoopEntry] != 65535)
		{
			printf ("%s\n", sFileNames[iLoopEntry]);
			if (iLoopEntry <= 108)
			{
				if (strcmp (sFileNames[iLoopEntry], files[iLoopEntry][0]) != 0)
				{
					printf ("[ FAIL ] Files do not match: \"%s\" and \"%s\"!\n",
						sFileNames[iLoopEntry], files[iLoopEntry][0]);
					exit (EXIT_ERROR);
				}
			}
			snprintf (sFileName, MAX_FILENAME, "%s%s%s",
				DIR_FILES, SLASH, sFileNames[iLoopEntry]);
			LoadFileCom (iFdA, sFileNames[iLoopEntry]);
			if (strcmp (files[iLoopEntry][1], "0") == 0)
			{
				SaveFileCom (sFileName);
			} else {
				DecompressFile (sFileNames[iLoopEntry]);
				SaveFileDecom (sFileName);
			}
		}
	}
}
/*****************************************************************************/
void LoadFileToDecom (char *sFileName)
/*****************************************************************************/
{
	int iFd;
	int iRead;
	char sRead[1 + 2];
	int iEOF;

	iFd = open (sFileName, O_RDONLY|O_BINARY);
	iSizeDecom = 0;
	iEOF = 0;
	do {
		iRead = read (iFd, sRead, 1);
		switch (iRead)
		{
			case 0: iEOF = 1; break;
			default:
				sFileDecom[iSizeDecom] = sRead[0];
				iSizeDecom++;
				break;
		}
	} while (iEOF == 0);
	sFileDecom[iSizeDecom] = '\0';
}
/*****************************************************************************/
void ApplySkillIfNecessary (int iLoc)
/*****************************************************************************/
{
	if (arGuardLocations[iCurRoom] == iLoc)
	{
		arGuardSkills[iCurRoom] = iGuardSkill;
		iChanged++;
	}
}
/*****************************************************************************/
void ChangeCustom (int iAmount)
/*****************************************************************************/
{
	if (((iAmount > 0) && (iCustomBlock < 0xFF)) ||
		((iAmount < 0) && (iCustomBlock > 0x00)))
	{
		/*** Modify the custom tile. ***/
		iCustomBlock+=iAmount;
		if (iCustomBlock < 0x00) { iCustomBlock = 0x00; }
		if (iCustomBlock > 0xFF) { iCustomBlock = 0xFF; }

		PlaySound ("wav/plus_minus.wav");
	}
}
/*****************************************************************************/
int OnTile (void)
/*****************************************************************************/
{
	// TODO

	/*** prince ***/
	if (InArea (447, 401, 447 + 32, 401 + 61) == 1) { return (128); }
	if (InArea (481, 401, 481 + 32, 401 + 61) == 1) { return (129); }

	return (0);
}
/*****************************************************************************/
int IsDisabled (int iTile)
/*****************************************************************************/
{
	if (iTile == 0) { } // TODO
	return (0); // TODO
}
/*****************************************************************************/
